"""
traverse_action.py – Populate a traverse action instance in PyRAL
"""

# System
import logging
from typing import Set, Dict, List, NamedTuple, Callable
from enum import Enum

# Model Integration
from scrall.parse.visitor import PATH_a
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction

# XUML_Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.flow import Flow
from xuml_populate.exceptions.action_exceptions import (UndefinedRelationship, IncompletePath,
                                                        NoDestinationInPath, UndefinedClass,
                                                        RelationshipUnreachableFromClass, HopToUnreachableClass,
                                                        MissingTorPrefInAssociativeRel, NoSubclassInHop,
                                                        SubclassNotInGeneralization, PerspectiveNotDefined,
                                                        UndefinedAssociation, NeedPerspectiveOrClassToHop,
                                                        NeedPerspectiveToHop, UnexpectedClassOrPerspectiveInPath)
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap
from xuml_populate.populate.mmclass_nt import (Action_i, Traverse_Action_i, Path_i, Hop_i, Association_Class_Hop_i,
                                               Circular_Hop_i, Symmetric_Hop_i, Asymmetric_Circular_Hop_i,
                                               Ordinal_Hop_i, Straight_Hop_i,
                                               From_Asymmetric_Association_Class_Hop_i,
                                               From_Symmetric_Association_Class_Hop_i, To_Association_Class_Hop_i,
                                               Perspective_Hop_i, Generalization_Hop_i, To_Subclass_Hop_i,
                                               To_Superclass_Hop_i, Association_Hop_i)
from xuml_populate.populate.actions.hop_types import (AggregationType, Hop, SymmetricHop, AsymmetricCircularHop,
                                                      OrdinalHop, FromAsymAssocHop)

_logger = logging.getLogger(__name__)

# Transactions
tr_Traverse = "Traverse Action"


class TraverseAction:
    """
    Populate a Traverse Action given a parsed Path and an already populated input Instance Flow.

    Ensure that the Path is well-formed and consistent with the Domain's Class Model.

    Populate an output Instance Flow produced by this Traverse Action.
    """

    def __init__(self, input_instance_flow: Flow_ap, path: PATH_a, activity_data: Activity_ap):
        """
        Initialize the python attributes required to manage validation and population.

        :param input_instance_flow: This is the source instance flow where the path begins
        :param path: Parsed Scrall representing a Path
        :param activity_data:
        """
        # Save parameter values
        self.input_instance_flow = input_instance_flow
        self.path = path
        self.anum = activity_data.anum
        self.domain = activity_data.domain
        self.activity_path = activity_data.activity_path
        self.scrall_text = activity_data.scrall_text
        self.mult = input_instance_flow.max_mult  # Will be updated as the max mult of the current hop

        # Initialize tracking attributes
        self.path_index = 0
        self.name = None
        self.id = None
        self.dest_class = None  # End of path
        self.class_cursor = None
        self.rel_cursor = None
        self.hops = []
        self.action_id = None
        self.dest_fid = None

        self.output_flow = self.build_path()

    def build_path(self) -> Flow_ap:
        """
        Step through a path populating it along the way.

        :return: The Traverse Action ID and the output instance flow id, its Class Type name and its maximum instance
        multiplicity, 1 or M
        """
        self.hops = []
        self.class_cursor = self.input_instance_flow.tname  # Validation cursor is on this class now
        self.name = "/"  # The path text forms path name value

        # Verify adequate path length
        if len(self.path.hops) < 2:
            raise IncompletePath(self.path)
        # Path has at least 2 hop elements

        # Validate destination class at the end of the path
        terminal_hop = self.path.hops[-1]
        if type(terminal_hop).__name__ != 'N_a':
            # Destination class must a name
            raise NoDestinationInPath(self.path)
        self.dest_class = terminal_hop.name

        # Valdiate path continuity
        # Step through the path validating each relationship, phrase, and class
        # Ensure that each step is reachable on the class model
        self.path_index = 0
        while self.path_index < len(self.path.hops) - 1:
            hop = self.path.hops[self.path_index]

            if type(hop).__name__ == 'N_a':
                # This should be a perspective since flow names get eaten in the relationship hop handlers
                # and a path cannot begin with a class name
                # (any class name prefixing a path will have been processed to populate a labeled instance flow earlier)
                if not self.resolve_perspective(phrase=hop.name) and not self.resolve_ordinal_perspective(
                        perspective=hop.name):
                    raise UnexpectedClassOrPerspectiveInPath(name=hop.name, path=self.path)

            elif type(hop).__name__ == 'R_a':
                self.rel_cursor = hop.rnum
                self.name += self.rel_cursor + '/'
                # This is either an Association, Generalization, or Ordinal Relationship
                # Determine the type and call the corresponding hop populator

                # First we look for any References to or from the class cursor
                R = f"(From_class:<{self.class_cursor}> OR To_class:<{self.class_cursor}>), Rnum:<{hop.rnum}>, " \
                    f"Domain:<{self.domain}>"
                ref_r = Relation.restrict(db=mmdb, restriction=R, relation="Reference")
                if ref_r.body:
                    P = ('Ref', 'From_class', 'To_class')
                    refs = Relation.project(db=mmdb, attributes=P, svar_name='rhop').body

                    # Generalization
                    if refs[0]['Ref'] == 'G':
                        self.hop_generalization(refs=refs)
                    else:
                        self.hop_association(refs=refs)
                else:
                    # The perspective must be specified in the next hop
                    self.path_index += 1
                    self.resolve_ordinal_perspective(perspective=self.path.hops[self.path_index].name)

            self.path_index += 1

        if self.dest_class != self.class_cursor:
            # Path does not reach destination
            pass

        # Now we can populate the path
        self.populate()

        return Flow_ap(fid=self.dest_fid, content=Content.INSTANCE, tname=self.dest_class, max_mult=self.mult)


    def populate(self):
        """
        Populate the Traverse Statement, Path and all Hops
        """
        self.name = self.name.rstrip('/')  # Remove trailing '/' from the path name
        # Create a Traverse Action and Path
        Transaction.open(db=mmdb, name=tr_Traverse)
        _logger.info(f"OPEN > {mmdb}:{tr_Traverse}")
        self.action_id = Action.populate(tr=tr_Traverse, anum=self.anum, domain=self.domain, action_type="traverse")
        # Create the Traverse action destination flow (the output for R930)
        self.dest_fid = Flow.populate_instance_flow(cname=self.dest_class, anum=self.anum,
                                                    domain=self.domain, label=None,
                                                    single=True if self.mult == MaxMult.ONE else False).fid

        _logger.info(f"INSERT Traverse action output Flow: ["
                     f"{self.domain}:{self.dest_class}:{self.activity_path.split(':')[-1]}"
                     f":{self.dest_fid}]")
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Traverse_Action', tuples=[
            Traverse_Action_i(ID=self.action_id, Activity=self.anum, Domain=self.domain, Path=self.name,
                              Source_flow=self.input_instance_flow.fid, Destination_flow=self.dest_fid)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Path', tuples=[
            Path_i(Name=self.name, Domain=self.domain, Dest_class=self.dest_class)
        ])

        # Get the next action ID
        # Then process each hop
        for number, h in enumerate(self.hops, start=1):
            # Call method to populate this particular kind of type of hop (straight, gen, to assoc, etc)
            params = vars(h).copy()
            params.pop("hoptype")
            params["number"] = number
            h.hoptype(**params)
            # h.hoptype(number=number, to_class=h.to_class, rnum=h.rnum)
        Transaction.execute(db=mmdb, name=tr_Traverse)
        _logger.info(f"EXECUTED > {mmdb}:{tr_Traverse}")

    def validate_rel(self, rnum: str):
        rel = f"Rnum:<{rnum}>, Domain:<{self.domain}>"
        if not Relation.restrict(db=mmdb, restriction=rel, relation="Relationship").body:
            _logger.error(f"Undefined Rnum {rnum} in Domain {self.domain}")
            raise UndefinedRelationship(rnum=rnum, domain=self.domain)

    def ordinal_hop(cls, cname: str, ascending: bool):
        _logger.info("ACTION:Traverse - Populating an ordinal hop")

    def symmetric_hop(cls, cname: str):
        _logger.info("ACTION:Traverse - Populating a circular symmetric hop")

    def asymmetric_circular_hop(self, cname: str, side: str):
        _logger.info("ACTION:Traverse - Populating an asymmetric circular hop")

    def from_symmetric_association_class(self, *, number: int, run: str, rnum: str):
        _logger.info("ACTION:Traverse - Populating a from symmetric assoc class hop")

    def from_asymmetric_association_class(self, *, number: int, rnum: str, to_class: str, side: str):
        """
        :param side: Perspective side (T or P)
        :return:
        """
        _logger.info("ACTION:Traverse - Populating a From Asymmetric Assoc Class Hop")
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='From_Asymmetric_Association_Class_Hop', tuples=[
            From_Asymmetric_Association_Class_Hop_i(Number=number, Path=self.name, Domain=self.domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Association_Class_Hop', tuples=[
            Association_Class_Hop_i(Number=number, Path=self.name, Domain=self.domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Association_Hop', tuples=[
            Association_Hop_i(Number=number, Path=self.name, Domain=self.domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Hop', tuples=[
            Hop_i(Number=number, Path=self.name, Domain=self.domain, Rnum=rnum, Class_step=to_class)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Perspective_Hop', tuples=[
            Perspective_Hop_i(Number=number, Path=self.name, Domain=self.domain, Side=side, Rnum=rnum)
        ])

    def to_association_class(self, *, number: int, rnum: str, to_class: str):
        """

        :param number:
        :param rnum:
        :param to_class:
        :return:
        """
        _logger.info("ACTION:Traverse - Populating a To Association Class Hop")
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='To_Association_Class_Hop', tuples=[
            To_Association_Class_Hop_i(Number=number, Path=self.name, Domain=self.domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Association_Class_Hop', tuples=[
            Association_Class_Hop_i(Number=number, Path=self.name, Domain=self.domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Association_Hop', tuples=[
            Association_Hop_i(Number=number, Path=self.name, Domain=self.domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Hop', tuples=[
            Hop_i(Number=number, Path=self.name, Domain=self.domain, Rnum=rnum, Class_step=to_class)
        ])

    def straight_hop(self, *, number: int, rnum: str, to_class: str):
        """
        Populate an instance of Straight HopArgs

        :param number:  Value (1, 2, 3... ) establishing order within a Path, See Hop.Number in the class model
        :param to_class: Hop over to this class
        :param rnum: Across this association
        """
        _logger.info("ACTION:Traverse - Populating a straight hop")
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Straight_Hop', tuples=[
            Straight_Hop_i(Number=number, Path=self.name, Domain=self.domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Association_Hop', tuples=[
            Association_Hop_i(Number=number, Path=self.name, Domain=self.domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Traverse, relvar='Hop', tuples=[
            Hop_i(Number=number, Path=self.name, Domain=self.domain, Rnum=rnum, Class_step=to_class)
        ])

    def to_superclass_hop(self):
        _logger.info("ACTION:Traverse - Populating a to superclass hop")

    def to_subclass_hop(self, sub_class: str):
        _logger.info("ACTION:Traverse - Populating a to subclass hop")

    def is_assoc_class(self, cname: str, rnum: str) -> bool:
        """
        Returns true
        :param cname: Class to investigate
        :param rnum: Class participates in this association
        :return: True of the class is an association class formalizing the specified association
        """
        r = f"Class:<{cname}>, Rnum:<{rnum}>, Domain:<{self.domain}>"
        return bool(Relation.restrict(db=mmdb, restriction=r, relation="Association_Class").body)

    def is_reflexive(self, rnum: str) -> int:
        """
        Is this a reflexive association and, if so, how many perspectives does it have?
        An association with both a T and P perspective is an asymmetric association while
        an association with a single S perspective is a symmetric association

        :param rnum: The association rnum to inspect
        :return: Zero if non-reflexive, 1 if symmetric and 2 if assymmetric reflexive
        """
        # Get all perspectives defined on rnum
        R = f"Rnum:<{rnum}>, Domain:<{self.domain}>"
        perspectives = Relation.restrict(db=mmdb, restriction=R, relation="Perspective")
        if not perspectives.body:
            # Every association relationship defines at least one perspective
            raise UndefinedAssociation(rnum=rnum, domain=self.domain)
        vclasses = Relation.project(db=mmdb, attributes=('Viewed_class',)).body
        # Reflexive if there is both viewed classes are the same (only 1)
        # So, if reflexive, return 1 (S - Symmetric) or 2 (T,P - Assymetric), otherwise 0, non-reflexive
        return len(perspectives.body) if len(vclasses) == 1 else 0

    def reachable_classes(self, rnum: str) -> Set[str]:
        """
        Return a set of all classes reachable on the provided relationship

        :param rnum:
        :return:
        """
        reachable_classes = set()
        R = f"Rnum:<{rnum}>, Domain:<{self.domain}>"
        refs = Relation.restrict(db=mmdb, restriction=R, relation="Reference").body
        for ref in refs:
            reachable_classes.add(ref['To_class'])
            reachable_classes.add(ref['From_class'])
        return reachable_classes

    def resolve_ordinal_perspective(self, perspective: str) -> bool:
        # Search for ordinal rel with the supplied perspective
        # TODO: Update metamodel with two additional identifiers
        R = f"Ranked_class:<{self.class_cursor}>, Domain:<{self.domain}>, " \
            f"(Ascending_perspective:<{perspective}> OR Descending_perspective:<{perspective}>)"
        orel = Relation.restrict(db=mmdb, restriction=R, relation="Ordinal_Relationship").body[0]
        if not orel:
            return False
        self.rel_cursor = orel['Rnum']
        self.ordinal_hop(cname=self.class_cursor, ascending=orel['Ascending_perspective'] == perspective)
        return True

    def hop_generalization(self, refs: List[Dict[str, str]]):
        """
        Populate a Generalization HopArgs

        :param refs:
        :return:
        """
        # If hopping from a superclass, all subclass references will be provided as refs
        # otherwise we are hopping from one of the subclasses and only one ref from that subclass is provided
        # The to class for each ref must be the superclass, so we just grab the first (possibly only) ref
        super_class = refs[0]['To_class']
        if len(refs) > 1:
            # We are hopping from the super_class to a subclass
            P = ("From_class",)
            sub_tuples = Relation.project(db=mmdb, attributes=P, relation="rhop").body
            subclasses = {s['From_class'] for s in sub_tuples}
            # The subclass must be specified in the next hop
            self.path_index += 1
            next_hop = self.path.hops[self.path_index]
            if next_hop.name not in subclasses:
                raise NoSubclassInHop(superclass=super_class, rnum=self.rel_cursor, domain=self.domain)
            self.class_cursor = next_hop.name
            self.to_subclass_hop(sub_class=self.class_cursor)
            return
        else:
            # # Superclass to subclass
            self.class_cursor = super_class
            self.to_superclass_hop()
            return

    def hop_association(self, refs: List[Dict[str, str]]):
        """
        Populate hop across the association

        :param refs: A list of tuple references where the to or from class is the cursor class
        """
        # Single reference, R, T or P
        if len(refs) == 1:
            ref, from_class, to_class = map(refs[0].get, ('Ref', 'From_class', 'To_class'))
            if ref == 'R':
                if to_class == from_class:
                    # This must be an asymmetric cycle unconditional on both ends
                    # which means a perspective must be specified like: /R1/next
                    # So we need to assume that the next hop is a perspective.
                    # We advance to the next hop in the path and then resolve the perspective
                    # (If it isn't a perspective, an exception will be raised in the perspective resolveer)
                    self.path_index += 1
                    self.resolve_perspective(phrase=self.path.hops[self.path_index])
                else:
                    # Add a straight hop to the hop list and update the class_cursor to either the to or from class
                    # whichever does not match the class_cursor

                    # Update multiplicity for this hop
                    # We need to look up the Perspective to get the multiplicity
                    # Since this is an R ref (no association class) we just need to specify the
                    # rnum, domain, and viewed class which will be the updated class cursor
                    self.class_cursor = to_class if to_class != self.class_cursor else from_class
                    R = f"Rnum:<{self.rel_cursor}>, Domain:<{self.domain}>, Viewed_class:<{self.class_cursor}>"
                    persp_r = Relation.restrict(db=mmdb, relation='Perspective', restriction=R)
                    if not persp_r.body:
                        # TODO: raise exception
                        return False
                    self.mult = MaxMult.ONE if persp_r.body[0]['Multiplicity'] == '1' else MaxMult.MANY
                    self.hops.append(
                        Hop(hoptype=self.straight_hop, to_class=self.class_cursor, rnum=self.rel_cursor))
                return

            if ref == 'T' or ref == 'P':
                # We are traversing an associative relationship
                # This means that we could be traversing to either the association class
                # or a straight hop to a participating class
                # We already know the association class as the from class. So we need to get both
                # to classes (same class may be on each side in a reflexive)
                # Then we look ahead for the next step which MUST be either a class name
                # or a perspective
                self.path_index += 1
                next_hop = self.path.hops[self.path_index]
                # The next hop must be either a class name or a perspective phrase on the current rel
                if type(next_hop).__name__ == 'R_a':
                    # In other words, it cannot be an rnum
                    raise NeedPerspectiveOrClassToHop(rnum=self.rel_cursor, domain=self.domain)
                # Is the next hop the association class?
                if next_hop.name == from_class:
                    self.class_cursor = from_class
                    # Update multiplicity
                    # First check multiplicity on to_class perspective (same as ref)
                    R = f"Rnum:<{self.rel_cursor}>, Domain:<{self.domain}>, Side:<{ref}>"
                    persp_r = Relation.restrict(db=mmdb, relation='Perspective', restriction=R)
                    if not persp_r.body:
                        # TODO: raise exception
                        return False
                    # Set multiplicity based on the perspective
                    self.mult = MaxMult.ONE if persp_r.body[0]['Multiplicity'] == '1' else MaxMult.MANY
                    # If multiplicity has been set to 1, but associative multiplicty is M, we need to set it as M
                    R = f"Rnum:<{self.rel_cursor}>, Domain:<{self.domain}>, Class:<{self.class_cursor}>"
                    assoc_class_r = Relation.restrict(db=mmdb, relation='Association_Class', restriction=R)
                    if not assoc_class_r.body:
                        # TODO: raise exception
                        return False
                    associative_mult = assoc_class_r.body[0]['Multiplicity']
                    # Associative mult of M overrides a single mult
                    self.mult = MaxMult.MANY if associative_mult == 'M' else self.mult

                    self.name += self.class_cursor + '/'
                    self.hops.append(
                        Hop(hoptype=self.to_association_class, to_class=self.class_cursor, rnum=self.rel_cursor)
                    )
                    return
                elif next_hop.name == to_class:
                    # Asymmetric reflexive hop requires a perspective phrase
                    raise NeedPerspectiveToHop(rnum=self.rel_cursor, domain=self.domain)

                else:
                    # Get the To class of the other (T or P) reference
                    other_ref_name = 'P' if ref == 'T' else 'T'
                    R = f"Ref:<{other_ref_name}>, Rnum:<{self.rel_cursor}>, Domain:<{self.domain}>"
                    other_ref = Relation.restrict(restriction=R, relation="Reference").body
                    if not other_ref:
                        # The model must be currupted somehow
                        raise MissingTorPrefInAssociativeRel(rnum=self.rel_cursor, domain=self.domain)
                    other_participating_class = other_ref[0]['To_class']
                    if next_hop.name == other_participating_class:
                        self.class_cursor = next_hop.name
                        self.straight_hop()
                        return
                    else:
                        # Next hop must be a perspective
                        self.resolve_perspective(phrase=next_hop.name)
                        return

        # T and P reference
        else:
            # Current hop is from an association class
            self.path_index += 1
            next_hop = self.path.hops[self.path_index]
            # Does the next hop match either of the participating classes
            particip_classes = {refs[0]['To_class'], refs[1]['To_class']}
            if next_hop.name in particip_classes:
                # The particpating class is explicitly named
                self.class_cursor = next_hop.name
                R = f"Viewed_class:<{self.class_cursor}>, Rnum:<{self.rel_cursor}>, Domain:<{self.domain}>"
                Relation.restrict(db=mmdb, relation='Perspective', restriction=R)
                P = ('Side',)
                side = Relation.project(db=mmdb, attributes=P).body[0]['Side']
                self.hops.append(
                    FromAsymAssocHop(hoptype=self.from_asymmetric_association_class, to_class=self.class_cursor,
                                     rnum=self.rel_cursor, side=side)
                )
                return
            else:
                # The next hop needs to be a perspective
                self.resolve_perspective(phrase=next_hop.name)
                return

    def resolve_perspective(self, phrase: str) -> bool:
        """
        Populate hop across the association perspective

        :param phrase:  Perspective phrase text such as 'travels along'
        """
        # Find phrase and ensure that it is on an association that involves the class cursor
        R = f"Phrase:<{phrase}>, Domain:<{self.domain}>"
        r_result = Relation.restrict(db=mmdb, relation='Perspective', restriction=R)
        if not r_result.body:
            return False
        P = ('Side', 'Rnum', 'Viewed_class')
        p_result = Relation.project(db=mmdb, attributes=P)
        side, rnum, viewed_class = map(p_result.body[0].get, P)
        self.rel_cursor = rnum

        # The next hop may be a class name that matches the viewed class
        # If so, we can move the path index forward so that we don't process that class as a separate hop
        try:
            next_hop = self.path.hops[self.path_index + 1]
            if next_hop.name == viewed_class:
                self.path_index += 1
        except (IndexError, AttributeError) as e:
            # We're already processing the last hop in the path, so don't bother advancing the path index or the
            # next hop is an rnum and not a name in which case we certainly don't want to advance the path index
            pass

        # We found the perspective
        # Now we decide which kind of hop to populate
        # We start by asking, "Is this association reflexive?"
        if symmetry := self.is_reflexive(rnum):
            # Symmetry is zero if non-reflexive, otherwise 1:symmetric, 2:asymmetric
            # So it must be either 1 or 2
            if self.class_cursor == viewed_class:
                # The class_cursor is one of the participating classes, i.e. not the association class
                # So it is a Circular HopArgs from-to the same class
                if symmetry == 1:
                    # Populate a symmetric hop
                    self.symmetric_hop(viewed_class)
                else:
                    # Populate an asymmetric hop
                    self.asymmetric_circular_hop(cname=viewed_class, side=side)
                return True  # Circular hop populated
            else:
                # The class_cursor must be the association class
                if symmetry == 1:
                    self.from_symmetric_association_class(rnum=rnum)
                else:
                    self.from_asymmetric_association_class(side=side)
                return True  # From assoc class hop populated
        else:  # Non-reflexive association (non-circular hop)
            # We are either hopping from the association class to a viewed class or
            # from the other participating class to the viewed class
            if self.is_assoc_class(cname=self.class_cursor, rnum=rnum):
                self.from_asymmetric_association_class(side=side)
            else:
                self.class_cursor = viewed_class
                # TODO: Supply params below
                self.straight_hop()
            return True  # Non-reflexive hop to a participating class



"""
ns_flow.py – Process a Non Scalar Flow
"""

import logging
from xuml_populate.config import mmdb
from pyral.relation import Relation
from typing import Optional, List, Dict
from xuml_populate.exceptions.action_exceptions import FlowException, NonScalarFlowRequired
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content
from xuml_populate.populate.mm_class import MMclass
from xuml_populate.populate.actions.table import Table

# TODO: Add Table and Control Flow population

_logger = logging.getLogger(__name__)


class NonScalarFlow:
    """
    Support queries common to Instance and Table Flows
    """

    domain = None
    activity = None
    label = None

    @classmethod
    def header_union(cls, a_flow: Flow_ap, b_flow: Flow_ap, domain: str) -> Dict[str, str]:
        # Create a list of headers, one per ns_flow
        headers = []
        for f in [a_flow, b_flow]:
            if f.content == Content.INSTANCE:
                headers.append(MMclass.header(cname=f.tname, domain=domain))
            else:
                headers.append(Table.header(tname=f.tname, domain=domain))
        # Convert the two headers from a list of dictionaries to a pair of tuples
        h = [(tuple(h.items())) for h in headers]
        # Union is formed by stepping through each attribute and adding it to a set
        u = {e for i in h for e in i}
        # Now convert back to a dictionary
        return {t[0]: t[1] for t in u}

    @classmethod
    def header(cls, ns_flow: Flow_ap, domain:str) -> Dict[str, str]:
        """
        Given a Non Scalar Flow, obtain its header

        :param ns_flow:
        :param domain:
        :return: Header as a dictionary of attr:scalar (type) pairs
        """
        match ns_flow.content:
            case Content.INSTANCE:
                return MMclass.header(ns_flow.tname, domain)
            case Content.RELATION:
                return Table.header(ns_flow.tname, domain)
            case _:
                raise NonScalarFlowRequired

    @classmethod
    def headers_disjoint(cls, a_flow: Flow_ap, b_flow: Flow_ap, domain: str) -> bool:
        """
        Ensure that there are no common attributes shared by more than on header

        :param a_flow:  A input
        :param b_flow:  B input
        :param domain: The domain name
        :return: True if A and B inputs are disjoint
        """
        # Create a list of headers, one per ns_flow
        headers = []
        for f in [a_flow, b_flow]:
            if f.content == Content.INSTANCE:
                headers.append(MMclass.header(cname=f.tname, domain=domain))
            else:
                headers.append(Table.header(tname=f.tname, domain=domain))
        # Convert the two headers from a list of dictionaries to a pair of tuples
        a_header, b_header = [(tuple(h.items())) for h in headers]
        # Check to see that the a and b headers in the list are disjoint
        return set(a_header).isdisjoint(b_header)

    @classmethod
    def same_headers(cls, a_flow: Flow_ap, b_flow: Flow_ap, domain: str) -> bool:
        """
        Ensure that each ns_flow shares the same headers

        :param a_flow:
        :param b_flow:
        :param domain:
        :return: True if all headers share the same set of attr/type pairs
        """
        # Create a list of headers, one per ns_flow
        headers = []
        for f in [a_flow, b_flow]:
            if f.content == Content.INSTANCE:
                headers.append(MMclass.header(cname=f.tname, domain=domain))
            else:
                headers.append(Table.header(tname=f.tname, domain=domain))

        # We need to freeze the list before we can do a set operation on it
        # Convert headers from a list of dictionaries to a list of tuples to a tuple of tuples
        headers = tuple([(tuple(a.items())) for a in headers])
        # If all headers are the same, there will be only one set element
        return len(set(headers)) == 1



""" system.py – Process all modeled domains within a system """

# System
import logging
from pathlib import Path
from contextlib import redirect_stdout

# Model Integration
from xcm_parser.class_model_parser import ClassModelParser
from xsm_parser.state_model_parser import StateModelParser
from op2_parser.op_parser import OpParser
from mtd_parser.method_parser import MethodParser
from pyral.relvar import Relvar

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.domain import Domain
from xuml_populate.populate.mmclass_nt import System_i

_mmdb_fname = f"{mmdb}.ral"
_logger = logging.getLogger(__name__)


class System:
    """
    The command line specifies a package representing a System. The organization of this package is defined
    in the readme file.

    We need to descend through that package loading and parsing all the files for the entire system.

    We then proceed to populate each modeled domain in the repository.
    """

    def __init__(self, name: str, system_path: Path, parse_actions: bool = False, verbose: bool = False):
        """
        Parse and otherwise process the contents of each modeled domain in the system.
        Then populate the content of each domain into the metamodel database.

        :param system_path: The path to the system package
        :param parse_actions: If true, all action text is parsed and populated into the metamodel,
        otherwise it is just kept as text
        """
        _logger.info(f"Processing system: [{system_path}]")

        self.name = name
        self.parse_actions = parse_actions
        self.content = {}  # Parsed content for all files in the system package
        self.mmdb_path = Path(__file__).parent / "populate" / _mmdb_fname  # Path to the serialized repository db
        self.system_name = system_path.stem.title()
        self.verbose=verbose

        # Process each domain folder in the system package
        for domain_path in system_path.iterdir():
            # First make sure it is really a domain folder, or at least a folder
            # For example, on mac OS we sometimes trip on a .DS_Store file and, if so, we want to ignore it
            if domain_path.name.startswith('.'):
                _logger.warning(f"Path: {domain_path} is hidden -- skipping")
                continue

            if not domain_path.is_dir():
                _logger.warning(f"Path: {domain_path} is not a directory -- skipping")
                continue
            # File names may differ from the actual model element name due to case and delimiter differences
            # For example, the domain name `Elevator Management` may have the file name `elevator-management`
            # The domain name will be in the parsed content, but it is convenient to use the file names as keys
            # to organize our content dictionary since we these are immediately available
            domain_name = None  # Domain name is unknown until the class model is parsed
            _logger.info(f"Processing domain: [{domain_path}]")

            subsys_folders = [f for f in domain_path.iterdir() if f.is_dir()]
            for subsys_path in subsys_folders:

                # Process the class model for this subsystem
                # The class file name must match the subsystem folder name
                # Any other .xcm files will be ignored (only one class model recognized per subsystem)
                cm_file_name = subsys_path.stem + ".xcm"
                cm_path = subsys_path / "class-model" / cm_file_name
                _logger.info(f"Processing class model: [{cm_path}]")
                # Parse the class model
                cm_parse = ClassModelParser.parse_file(file_input=cm_path, debug=False)

                # If this is the first subsystem in the domain, get the domain name from the cm parse
                # domain will be None on the first subsystem
                if not domain_name:
                    domain_name = cm_parse.domain['name']
                    domain_alias = cm_parse.domain['alias']
                    # Create dictionary key for domain content
                    self.content[domain_name] = {'alias': domain_alias, 'subsystems': {}}

                # Get this subsystem name from the parse
                subsys_name = cm_parse.subsystem['name']

                # We add the subsystem dictionary to the system content for the current domain file name
                # inserting the class model parse
                self.content[domain_name]['subsystems'][subsys_name] = {
                    'class_model': cm_parse, 'methods': {}, 'state_models': {}, 'external': {}
                }

                # Load and parse all the methods for the current subsystem folder
                method_path = subsys_path / "methods"
                if method_path.is_dir():
                    # Find all class folders in the current subsystem methods directory
                    class_folders = [f for f in method_path.iterdir() if f.is_dir()]
                    for class_folder in class_folders:
                        # Process each method file in this class folder
                        for method_file in class_folder.glob("*.mtd"):
                            method_name = method_file.stem
                            _logger.info(f"Processing method: [{method_file}]")
                            # Parse the method file and insert it in the subsystem subsys_parse
                            mtd_parse = MethodParser.parse_file(method_file, debug=False)
                            self.content[domain_name]['subsystems'][subsys_name]['methods'][method_name] = mtd_parse
                else:
                    _logger.info("No method dir")

                # Load and parse the current subsystem's state models (state machines)
                sm_path = subsys_path / "state-machines"
                if sm_path.is_dir():
                    for sm_file in sm_path.glob("*.xsm"):
                        sm_name = sm_file.stem
                        _logger.info(f"Processing state model: [{sm_file}]")
                        # Parse the state model
                        sm_parse = StateModelParser.parse_file(file_input=sm_file, debug=False)
                        self.content[domain_name]['subsystems'][subsys_name]['state_models'][sm_name] = sm_parse
                else:
                    _logger.info("No state-machines dir")

                # Load and parse the external entity operations
                ext_path = subsys_path / "external"
                if ext_path.is_dir():
                    for ee_path in ext_path.iterdir():
                        ee_name = ee_path.name
                        self.content[domain_name]['subsystems'][subsys_name]['external'][ee_name] = {}
                        for op_file in ee_path.glob("*.op"):
                            op_name = op_file.stem
                            _logger.info(f"Processing ee operation: [{op_file}]")
                            op_parse = OpParser.parse_file(file_input=op_file, debug=False)
                            self.content[domain_name]['subsystems'][subsys_name]['external'][ee_name][op_name] = op_parse
                else:
                    _logger.info("No external dir")

        self.populate()

    def populate(self):
        """Populate the database from the parsed input"""

        # Initiate a connection to the TclRAL database
        from pyral.database import Database  # Metamodel load or creates has already initialized the DB session
        _logger.info("Initializing TclRAL database connection")
        Database.open_session(mmdb)

        # Start with an empty metamodel repository
        _logger.info("Loading Blueprint MBSE metamodel repository schema")
        Database.load(db=mmdb, fname=str(self.mmdb_path))

        # Populate the single instance System class
        Relvar.insert(db=mmdb, relvar='System', tuples=[
            System_i(Name=self.system_name),
        ])

        # Populate each domain into the metamodel db
        for domain_name, domain_parse in self.content.items():
            Domain(domain=domain_name, content=domain_parse, parse_actions=self.parse_actions, verbose=self.verbose)

        # Save the populated metamodel
        saved_mmdb_name = f"mmdb_{self.name}.ral"
        Database.save(db=mmdb, fname=saved_mmdb_name)

        # Output a text file of the populated mmdb
        mmdb_printout = f"mmdb_{self.name}.txt"
        with open(mmdb_printout, 'w') as f:
            with redirect_stdout(f):
                Relvar.printall(db=mmdb)



"""
attribute.py – Process parsed attribute to populate the metamodel db
"""

import logging
from xuml_populate.config import mmdb
from typing import Set
from pyral.relvar import Relvar
from pyral.relation import Relation
from xuml_populate.exceptions.action_exceptions import UndefinedAttribute
from xuml_populate.populate.mm_type import MMtype
from xuml_populate.populate.mmclass_nt import \
    Attribute_i, Non_Derived_Attribute_i, Model_Attribute_i, \
    Identifier_i, Irreducible_Identifier_i, Super_Identifier_i, Identifier_Attribute_i

UNRESOLVED = '__unresolved__'  # Attribute type is referential and type resolution is deferred

_logger = logging.getLogger(__name__)


class Attribute:
    """
    Populate all relevant Attribute relvars
    """

    record = None
    dtype = None
    participating_ids = None

    @classmethod
    def scalar(cls, name: str, tname: str, domain: str) -> str:
        """
        Returns the name of the Scalar (Type) of the specified Attribute.

        Raises an exception if the Attribute is not defined on the specified Class.

        :param name:  Attribute name
        :param tname:  Table (possibly Class) name
        :param domain:  Domain name
        :return: Name of Attribute's Scalar (Type)
        """
        R = f"Name:<{name}>, Non_scalar_type:<{tname}>, Domain:<{domain}>"
        model_attr_r = Relation.restrict(db=mmdb, relation='Model Attribute', restriction=R, svar_name="model_attr_rv")
        if not model_attr_r.body:
            _logger.error(f"Undefined attribute: [{name}:{tname}:{domain}]")
            raise UndefinedAttribute
        class_attr_r = Relation.semijoin(db=mmdb, rname1="model_attr_rv", rname2="Attribute",
                                         attrs={"Name":"Name", "Non_scalar_type":"Class", "Domain":"Domain"})
        if class_attr_r.body:
            return class_attr_r.body[0]['Scalar']

        table_attr_r = Relation.semijoin(db=mmdb, rname1="model_attr_rv", rname2="Table_Attribute",
                                         attrs={"Name":"Name", "Non_scalar_type":"Table", "Domain":"Domain"})
        if table_attr_r.body:
            return table_attr_r.body[0]['Scalar']


    @classmethod
    def populate(cls, tr: str, domain: str, cname: str, class_identifiers: Set[int], record):
        """
        Populate all relevant Attribute relvars

        :param tr: The open db transaction name
        :param domain: The domain name
        :param cname: The class name
        :param class_identifiers: The set of class identifier numbers
        :param record: The parsed class information
        """

        cls.record = record
        cls.dtype = record.get('type', UNRESOLVED)
        participating_ids = cls.record.get('I', [])  # This attr might not participate in any identifier
        # Populate the Attribute's type if it hasn't already been populated
        MMtype.populate_unknown(name=cls.dtype, domain=domain)
        Relvar.insert(mmdb, tr=tr, relvar='Attribute', tuples=[
            Attribute_i(Name=record['name'], Class=cname, Domain=domain, Scalar=cls.dtype)
        ])
        Relvar.insert(mmdb, tr=tr, relvar='Model_Attribute', tuples=[
            Model_Attribute_i(Name=record['name'], Domain=domain, Non_scalar_type=cname)
        ])
        # TODO: Check for derived or non-derived, for now assume the latter
        Relvar.insert(mmdb, tr=tr, relvar='Non_Derived_Attribute', tuples=[
            Non_Derived_Attribute_i(Name=record['name'], Class=cname, Domain=domain)
        ])

        for i in participating_ids:
            # Add Identifier if it is not already in the population
            if i.number not in class_identifiers:
                Relvar.insert(mmdb, tr=tr, relvar='Identifier', tuples=[
                    Identifier_i(Number=i.number, Class=cname, Domain=domain)
                ])
                if not i.superid:
                    Relvar.insert(mmdb, tr=tr, relvar='Irreducible_Identifier', tuples=[
                        Irreducible_Identifier_i(Number=i.number, Class=cname, Domain=domain)
                    ])
                else:
                    Relvar.insert(mmdb, tr=tr, relvar='Super_Identifier', tuples=[
                        Super_Identifier_i(Number=i.number, Class=cname, Domain=domain)
                    ])
                class_identifiers.add(i.number)

            # Include this attribute in this identifier
            Relvar.insert(mmdb, tr=tr, relvar='Identifier_Attribute', tuples=[
                Identifier_Attribute_i(Identifier=i.number, Attribute=record['name'], Class=cname, Domain=domain)
            ])

    @classmethod
    def ResolveAttrTypes(cls, domain: str):
        """
        Determine an update type of each unresolved (referential) attribute
        """
        # TODO: Make this first part work and deprecate second
        R = f"Scalar:<{UNRESOLVED}>, Domain:<{domain}>"
        Relation.restrict(mmdb, relation='Attribute', restriction=R)
        uattrs = Relation.project(mmdb, attributes=('Name', 'Class'))

        # Rather than batch all the updates, we do them one by one
        # This reduces the search space for each subsequent type resolution
        for a in uattrs.body:
            assign_type = cls.ResolveAttr(attr=a['Name'], cname=a['Class'], domain=domain)
            Relvar.updateone(mmdb, relvar_name='Attribute',
                             id={'Name': a['Name'], 'Class': a['Class'], 'Domain': domain},
                             update={'Scalar': assign_type})

        # All attr types resolved, so delete the dummy UNRESOLVED type
        MMtype.depopulate_scalar(name=UNRESOLVED, domain=domain)

    @classmethod
    def ResolveAttr(cls, attr: str, cname: str, domain: str) -> str:
        """
        The modeler specifies explicit types only for non-referential attributes.
        This means that all attributes with unresolved types are referential.

        We need to obtain one (there could be multiple) Attribute Reference where the unresolved attribute is a source
        *From attribute* referring to some *To attribute*. Then we check the type of that *To attribute*.

        If the type is not <unresolved>, we return it. Otherwise, we recursively apply the same process to
        the *To attribute*.

        The chain of references must eventually land on a specified type if the model has been properly formalized.

        :param attr: Unresolved attribute: A referential attribute with an unresolved type
        :param cname: The class name
        :param domain: The domain name
        :return: Type name to assign
        """
        _logger.info(f"Resolving attribute type [{cname}.{attr}]")
        # We join the two relvars on the To_attribute so that we can obtain that attribute's Type

        Relation.join(mmdb, rname1='Attribute', rname2='Attribute_Reference',
                      attrs={'Name': 'To_attribute', 'Class': 'To_class', 'Domain': 'Domain'})

        # Finally, we restrict and project on our from attribute to get its reference type
        R = f"From_attribute:<{attr}>, From_class:<{cname}>, Domain:<{domain}>"
        from_attrs = Relation.restrict(mmdb, restriction=R)

        # The same attribute could participate in multiple References, so we just pick one arbitrarily
        aref = from_attrs.body[0]
        to_name, to_class, to_type = aref['Name'], aref['Class'], aref['Scalar']

        if to_type != UNRESOLVED:
            return to_type  # The To_attribute has a type
        else:
            # The To_attribute is also unresolved. Resolve it!
            return cls.ResolveAttr(attr=to_name, cname=to_class, domain=domain)



"""
instance_assignment.py – Break an instance set generator into one or more components
"""

import logging
from typing import Set, Dict, List, Optional
from xuml_populate.config import mmdb
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.mmclass_nt import Labeled_Flow_i
from xuml_populate.populate.actions.expressions.instance_set import InstanceSet
from xuml_populate.exceptions.action_exceptions import AssignZeroOneInstanceHasMultiple
from xuml_populate.populate.actions.aparse_types import (Flow_ap, MaxMult, Content, Activity_ap, Boundary_Actions,
                                                         Labeled_Flow)
from scrall.parse.visitor import Inst_Assignment_a

from pyral.transaction import Transaction
from pyral.relation import Relation  # For debugging
from pyral.relvar import Relvar

_logger = logging.getLogger(__name__)

# Transactions
tr_Migrate = "Migrate to Label"

class InstanceAssignment:
    """
    Break down a Scrall instance assignment statement into action semantics and populate them

    The lhs (left hand side) will be a labeled Instance Flow. It may or may not have an explicit Class Type.
    The card (cardinality) is either 1c or Mc (one or many conditional). It determines whether the lhs Instance
    Flow is Single or Multiple.

    The rhs (right hand side) is an expression that outputs an instance set of some Class Type. If the lhs is
    explicitly typed, we throw an exception if the rhs and lhs types do not match.

    For now we limit the expression to a chain of the following components:
        * create action
        * traversal action
        * instance flow
        * method or operation output of Class Type
        * selection output

    We say 'chain' since the output of one can feed into the input of the next yielding a final output at the end
    of the chain. It is this final output that determines the type (or type conflict) with the lhs Instance Flow

    We say 'for now' because this chain does not yet take into account instance set operations (add, subtract, union,
    etc). The Scrall syntax will later be udpated to accommodate such expressions.
    """

    assign_zero_one = None  # Does assignment operator limit to a zero or one instance selection?

    @classmethod
    def process(cls, activity_data: Activity_ap, inst_assign: Inst_Assignment_a,
                case_name: str, case_outputs: Set[Labeled_Flow]) -> Boundary_Actions:
        """
        Given a parsed instance set expression, populate each component action
        and return the resultant Class Type name

        We'll need an initial flow and we'll need to create intermediate instance flows to connect the components.
        The final output flow must be an instance flow. The associated Class Type determines the type of the
        assignment which must match any explicit type.

        :param case_outputs:
        :param case_name:
        :param inst_assign: The instance assignment statement parse
        :param activity_data: The enveloping anum
        :param case_prefix:
        """
        lhs = inst_assign.lhs  # Left hand side of assignment
        assign_zero_one = True if inst_assign.card == '1' else False  # Do we flow out one or many instances?
        rhs = inst_assign.rhs  # Right hand sid of assignment

        # The executing instance is by nature a single instance flow
        xi_instance_flow = Flow_ap(fid=activity_data.xiflow, content=Content.INSTANCE, tname=activity_data.cname,
                                   max_mult=MaxMult.ONE)

        # Process the instance set expression in the RHS and obtain the generated instance flow
        initial_aid, final_aid, iset_instance_flow = InstanceSet.process(input_instance_flow=xi_instance_flow,
                                                                         iset_components=rhs.components,
                                                                         activity_data=activity_data)

        # Process LHS after all components have been processed
        if assign_zero_one and iset_instance_flow.max_mult == MaxMult.ONE:
            raise AssignZeroOneInstanceHasMultiple(path=activity_data.activity_path, text=activity_data.scrall_text,
                                                   x=inst_assign.X)

        case_prefix = '' if not case_name else f"{case_name}_"
        output_flow_label = case_prefix + lhs.name.name
        if case_name:
            case_outputs.add(Labeled_Flow(label=output_flow_label, flow=iset_instance_flow))
        if lhs.exp_type and lhs.exp_type != iset_instance_flow.tname:
            # Raise assignment type mismatch exception
            pass

        # Migrate the RHS output to a labeled flow using the output flow label
        Transaction.open(db=mmdb, name=tr_Migrate)  # LHS labeled instance flow

        # Delete the Unlabeled flow
        Relvar.deleteone(db=mmdb, tr=tr_Migrate, relvar_name="Unlabeled_Flow",
                         tid={"ID": iset_instance_flow.fid, "Activity": activity_data.anum,
                              "Domain": activity_data.domain})
        # Insert the labeled flow
        Relvar.insert(db=mmdb, tr=tr_Migrate, relvar='Labeled_Flow', tuples=[
            Labeled_Flow_i(ID=iset_instance_flow.fid, Activity=activity_data.anum, Domain=activity_data.domain,
                           Name=output_flow_label)
        ])

        Transaction.execute(db=mmdb, name=tr_Migrate)  # LHS labeled instance flow
        pass

        return Boundary_Actions(ain={initial_aid}, aout={final_aid})



"""
parameter_validation.py – Verify that a parameter is defined
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.exceptions.action_exceptions import UndefinedParameter, UndefinedMethod
from xuml_populate.populate.actions.aparse_types import Activity_ap
from pyral.relation import Relation

_logger = logging.getLogger(__name__)

def validate_param(name: str, activity: Activity_ap):
    """
    Raise an exception if the specified Parameter is not defined

    :param name:  Parameter name
    :param activity:  Activity context
    """
    # TODO: Only works for Method Signature now. Add State and Operation Signatures later
    domain = activity.domain
    anum = activity.anum

    # Method signature
    if activity.cname and activity.opname:
        R = f"Method:<{activity.opname}>, Class:<{activity.cname}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Method_Signature', restriction=R)
        if not result.body:
            raise UndefinedMethod
        signum = result.body[0]['SIGnum']

        R = f"Name:<{name}>, Signature:<{signum}>, Activity:<{anum}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Parameter', restriction=R)
        if not result.body:
            raise UndefinedParameter
        return

    # State signature
    if activity.sname:
        # Fail until validation is added
        raise UndefinedParameter

    # Operation signature
    if activity.eename:
        # Fail until validation is added
        raise UndefinedParameter




""" table_expr.py -- Walk through a table expression and populate elements """

import logging
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.expressions.instance_set import InstanceSet
from xuml_populate.populate.actions.expressions.class_accessor import ClassAccessor
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.actions.rename_action import RenameAction
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap, Boundary_Actions
from xuml_populate.populate.actions.select_action import SelectAction
from xuml_populate.populate.actions.restrict_action import RestrictAction
from xuml_populate.populate.actions.project_action import ProjectAction
from xuml_populate.populate.actions.set_action import SetAction
from xuml_populate.exceptions.action_exceptions import (ActionException, TableOperationOrExpressionExpected,
                                                        FlowException,
                                                        UndefinedHeaderExpressionOp)
from scrall.parse.visitor import TEXPR_a
from pyral.relation import Relation

_logger = logging.getLogger(__name__)


class TableExpr:
    """
    For reference, a table expression is defined in the scrall grammar as:

    table_expr = table_operation
    table_operation = table_term (TOP table_term)*
    table_term = table / "(" table_expr ")" header_expr? selection? projection?
    TOP = '^' / '+' / '-' / '*' / '%' / '##'
    table = instance_set header_expr? selection? projection?

    (whitespace tokens removed in the above extract)

    So we need to walk through the parse tree through the nested operations, possibly
    building instance sets and then handling various header expressions (rename, etc),
    selections, and projections. These are all delegated elsewhere.

    Here we focus on unnesting all those table ops and operands.
    """
    tuple_output = None
    text = None  # A text representation of the expression
    domain = None
    anum = None
    scrall_text = None
    activity_path = None
    component_flow = None
    output_tflow_id = None
    activity_data = None
    action_inputs = None
    action_outputs = None

    @classmethod
    def process(cls, tuple_output: bool, activity_data: Activity_ap, rhs: TEXPR_a, input_instance_flow: Flow_ap) -> (
            Boundary_Actions, Flow_ap):
        """
        Initiate the recursive descent through a table expression and yield a final result as either a
        Table or Tuple Flow.

        :param tuple_output: The final output is a Tuple Flow
        :param activity_data: Context of the enclosing activity
        :param rhs: The right hand side of a table assignment
        :param input_instance_flow: This instance flow is available
        :return: boundary actions and a flow summary
        """
        cls.activity_data = activity_data
        cls.domain = activity_data.domain
        cls.anum = activity_data.anum
        cls.activity_path = activity_data.activity_path
        cls.scrall_text = activity_data.scrall_text

        cls.action_outputs = {}  # ID's of all Action output Data Flows
        cls.action_inputs = {}  # ID's of all Action input Data Flows

        final_output_flow = cls.walk(texpr=rhs, input_nsflow=input_instance_flow)
        # If tuple output has been specified, for a tuple assignment for example, tighten the max multiplicity to one

        all_ins = {v for s in cls.action_inputs.values() for v in s}
        all_outs = {v for s in cls.action_outputs.values() for v in s}
        init_aids = {a for a in cls.action_inputs.keys() if not cls.action_inputs[a].intersection(all_outs)}
        final_aids = {a for a in cls.action_outputs.keys() if not cls.action_outputs[a].intersection(all_ins)}

        return Boundary_Actions(ain=init_aids, aout=final_aids), final_output_flow

    @classmethod
    def walk(cls, texpr: TEXPR_a, input_nsflow: Flow_ap) -> Flow_ap:
        """

        :param input_nsflow:
        :param texpr: Parsed table expression
        """
        # Process the table component
        # It is either an instance set, name, or a nested table operation
        component_flow = input_nsflow
        match type(texpr.table).__name__:
            case 'N_a' | 'IN_a':
                # Is the name an existing Labeled Flow?
                R = f"Name:<{texpr.table.name}>, Activity:<{cls.anum}>, Domain:<{cls.domain}>"
                result = Relation.restrict(db=mmdb, relation='Labeled_Flow', restriction=R)
                if result.body:
                    # Name corresponds to some Labeled Flow instance
                    label_fid = result.body[0]['ID']
                    component_flow = Flow.lookup_data(fid=label_fid, anum=cls.anum, domain=cls.domain)
                else:
                    # Not a Labled Flow instance
                    # Is it a class name?  If so, we'll need a Class Accessor populated if we don't have one already
                    class_flow = ClassAccessor.populate(name=texpr.table.name, anum=cls.anum, domain=cls.domain)
                    if class_flow:
                        # We have a Class Accessor either previously or just now populated
                        # Set its output flow to the current component output
                        component_flow = Flow_ap(fid=class_flow, content=Content.INSTANCE,
                                                 tname=texpr.table.name, max_mult=MaxMult.MANY)
                    else:
                        # Neither labeled flow or class
                        # TODO: check for other possible cases
                        _logger.error(f"Name [{texpr.table.name}] does not label any flow")
                        raise FlowException

            case 'INST_a':
                # Process the instance set and obtain its flow id
                input_flow = component_flow
                initial_aid, final_aid, component_flow = InstanceSet.process(input_instance_flow=input_flow,
                                                                             iset_components=texpr.table.components,
                                                                             activity_data=cls.activity_data)
                # Add the output flow generated by the instance set expression to the set of ouput flows
                if initial_aid:
                    # For an InstanceSet with a single labeled flow component, no action is created
                    # So don't process action inputs and outputs unless there is an initial_aid
                    cls.action_inputs[initial_aid] = {input_flow.fid}
                    if final_aid:
                        cls.action_outputs[final_aid] = {component_flow.fid}
            case 'TOP_a':
                # The table is an operation on one or more operands
                # We need to process each operand
                # text = f" {texpr.op} "  # Flatten operator into temporary string
                # insert Computation and set its operator attribute with texpr.op
                operand_flows = []
                for o in texpr.table.operands:
                    operand_flows.append(cls.walk(texpr=o, input_nsflow=component_flow))
                op_name = texpr.table.op
                aid, component_flow = SetAction.populate(a_input=operand_flows[0], b_input=operand_flows[1],
                                                         setop=op_name, activity_data=cls.activity_data)
                cls.action_inputs[aid] = {operand_flows[0].fid, operand_flows[1].fid}
                cls.action_outputs[aid] = {component_flow.fid}
                pass
            case _:
                _logger.error(
                    f"Expected INST, N, IN or TOP, but received {type(texpr).__name__} during table_expr walk")
                raise TableOperationOrExpressionExpected
        # Process optional header, selection, and projection actions for the TEXPR
        if texpr.hexpr:
            for header_op in texpr.hexpr:
                # Process each header operation
                match type(header_op).__name__:
                    case 'Rename_a':
                        # Populate a rename relational action
                        input_flow = component_flow
                        aid, component_flow = RenameAction.populate(input_nsflow=input_flow,
                                                                    from_attr=header_op.from_name,
                                                                    to_attr=header_op.to_name,
                                                                    activity_data=cls.activity_data)
                        cls.action_inputs[aid] = {input_flow.fid}
                        cls.action_outputs[aid] = {component_flow.fid}
                    case 'Extend':
                        pass
                        # print()
                    case _:
                        raise UndefinedHeaderExpressionOp
                pass
        if texpr.selection:
            # If there is a selection on the instance set, create the action and obtain its flow id
            input_flow = component_flow  # TODO: input/component_flow redundant, just use component_flow
            if input_flow.content == Content.RELATION:
                aid, component_flow, input_sflows = RestrictAction.populate(input_relation_flow=input_flow,
                                                                            selection_parse=texpr.selection,
                                                                            activity_data=cls.activity_data)
                cls.action_inputs[aid] = {input_flow.fid}.union({f.fid for f in input_sflows})
                cls.action_outputs[aid] = {component_flow.fid}
            elif input_flow.content == Content.INSTANCE:
                select_action = SelectAction(input_instance_flow=input_flow, selection_parse=texpr.selection,
                                             activity_data=cls.activity_data)
                aid = select_action.action_id
                component_flow = select_action.output_instance_flow
                input_sflows = select_action.sflows
                cls.action_inputs[aid] = {input_flow.fid}.union({f.fid for f in input_sflows})
                cls.action_outputs[aid] = {component_flow.fid}
            else:
                raise ActionException
        if texpr.projection:
            # If there is a projection, create the action and obtain its flow id
            input_flow = component_flow
            aid, component_flow = ProjectAction.populate(input_nsflow=input_flow,
                                                         projection=texpr.projection, activity_data=cls.activity_data)
            cls.action_inputs[aid] = {input_flow.fid}
            cls.action_outputs[aid] = {component_flow.fid}
        return component_flow



""" test_parse.py -- Test successful parsing of *.op files """

import pytest
from pathlib import Path
from op2_parser.op_parser import OpParser

operations = [
    "arrived-at-floor",
    "goto-floor",
]

@pytest.mark.parametrize("op", operations)
def test_ops_pdf(op):

    input_path = Path(__file__).parent / "operations" / f"{op}.op"
    result = OpParser.parse_file(file_input=input_path, debug=False)
    assert result



"""
ee.py – Convert external entity to a relation
"""

import logging
from xuml_populate.config import mmdb
from pyral.transaction import Transaction
from pyral.relvar import Relvar
from xuml_populate.populate.element import Element
from xuml_populate.populate.operation import Operation
from xuml_populate.populate.mmclass_nt import External_Entity_i

_logger = logging.getLogger(__name__)

# Transactions
tr_EE = "EE"

class EE:
    """
    Populate all relevant External Entity relvars
    """
    subsys_ee_path = None
    record = None

    @classmethod
    def populate(cls, ee_name: str, subsys: str, domain: str, op_parse):
        """
        Populate an External Entity

        :param ee_name: The EE name
        :param cname: The proxy Class name
        :param subsys: The name of the subsystem
        :param domain: The name of the domain
        """
        # Class name can be obtained from the parse of any operation
        cname = next(iter(op_parse.values())).cname
        # Populate ee
        _logger.info(f"Populating ee [{ee_name}]")
        _logger.info(f"Transaction open: Populate EE")
        Transaction.open(mmdb, tr_EE)  # Create an EE with at least one Operation
        EEnum = Element.populate_unlabeled_subsys_element(tr=tr_EE, prefix='EE', subsystem=subsys, domain=domain)
        Relvar.insert(mmdb, tr=tr_EE, relvar='External_Entity', tuples=[
            External_Entity_i(EEnum=EEnum, Name=ee_name, Class=cname, Domain=domain)
        ])

        # Add operations
        tr = tr_EE  # Start out by completing the required op for the EE
        for op in op_parse.values():
            tr = Operation.populate(tr=tr, domain=domain, subsys=subsys, parsed_op=op)
            # "EE" (tr_EE) closes on first operation and all subsequent tr values will be "Operation"




"""
hop_types.py – Populate a traverse action instance in PyRAL
"""

# System
from dataclasses import dataclass
from typing import Callable
from enum import Enum

class AggregationType(Enum):
    FIRST = 1
    LAST = 2
    ALL = 3

# I tried NamedTuple and ran into type checking issues with inheritance, so let's try
# the dataclass decorator instead

@dataclass
class Hop:
    """
    Here we specify a method and the values it requires to populate the Hop subclasses
    in the metamodel navigation subsystem.
    This class is sufficient for Straight, From Symmetric Association Class, To Association Class and Generaization
    Hop subclasses.

    The subclass hop named tuples supply additional attributes as determined by R931, R936, R932, and R937
    of the metammodel navigation subsystem.
    """
    hoptype: Callable
    to_class: str
    rnum: str

@dataclass
class SymmetricHop(Hop):
    aggregation: AggregationType

@dataclass
class AsymmetricCircularHop(Hop):
    side: str
    aggregation: AggregationType

@dataclass
class OrdinalHop(Hop):
    side: str
    ascending: bool
    aggregation: AggregationType

@dataclass
class FromAsymAssocHop(Hop):
    side: str




"""
switch_action.py – Populate a switch action instance in PyRAL
"""

import logging
from typing import List
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.mm_class import MMclass
from scrall.parse.visitor import Projection_a
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.mmclass_nt import Read_Action_i, Attribute_Read_Access_i
from pyral.relvar import Relvar
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Switch = "Switch Action"

class SwitchAction:
    """
    Create all relations for a Switch Action
    """

    input_instance_flow = None  # We are selecting instances from this instance flow
    output_instance_flow = None
    expression = None
    comparison_criteria = []
    equivalence_criteria = []
    restriction_text = ""
    cardinality = None
    action_id = None
    criterion_ctr = 0
    max_mult = None

    @classmethod
    def populate(cls, input_single_instance_flow: Flow_ap, projection: Projection_a,
                 activity_data: Activity_ap) -> (str, List[Flow_ap]):
        """
        Populate the Switch Action

        :param input_single_instance_flow: The source flow into this selection
        :param projection:
        :param input_single_instance_flow:
        :param activity_data:
        :return: A list of scalar flows in the order of the project statement
        """
        anum = activity_data.anum
        domain = activity_data.domain
        si_flow = input_single_instance_flow  # Short name for convenience

        # Get the class header
        class_attrs = MMclass.header(cname=si_flow.tname, domain=domain)

        # Populate the Action superclass instance and obtain its action_id
        Transaction.open(db=mmdb, name=tr_Switch)
        action_id = Action.populate(tr=tr_Switch, anum=anum, domain=domain, action_type="switch")  # Transaction open
        Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Read_Action', tuples=[
            Read_Action_i(ID=action_id, Activity=anum, Domain=domain, Instance_flow=input_single_instance_flow.fid)
        ])
        scalar_flows = []
        proj_attrs = [n.name for n in projection.attrs] if projection.expand != 'ALL' else list(class_attrs.keys())
        for pa in proj_attrs:
            of = Flow.populate_scalar_flow(scalar_type=class_attrs[pa], anum=anum, domain=domain, label=None)
            Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Attribute_Read_Access', tuples=[
                Attribute_Read_Access_i(Attribute=pa, Class=si_flow.tname, Read_action=action_id, Activity=anum,
                                        Domain=domain, Output_flow=of.fid)
            ])
            scalar_flows.append(of)

            # output_flows[pa] = of
        # We now have a transaction with all select-action instances, enter into the metamodel db
        Transaction.execute(db=mmdb, name=tr_Switch)  # Select Action
        return action_id, scalar_flows



""" table_attribute.py - Provides table attribute functions """

import logging
from xuml_populate.config import mmdb
from xuml_populate.exceptions.action_exceptions import UndefinedTableAttribute
from pyral.relvar import Relvar
from pyral.relation import Relation

_logger = logging.getLogger(__name__)


class TableAttribute:
    """
    Populates the Table and all unconditionally related classes.
    This includes Table Attribute, Model Attribute and Table Flow
    """

    @classmethod
    def scalar(cls, name: str, table: str, domain: str) -> str:
        """
        Returns the name of the Scalar (Type) of the specified Attribute.

        Raises an exception if the Table Attribute is not defined on the specified Table.

        :param name:  Attribute name
        :param table: Table name
        :param domain:  Domain name
        :return: Name of Table Attribute's Scalar (Type)
        """
        R = f"Name:<{name}>, Table:<{table}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Table_Attribute', restriction=R)
        if not result.body:
            _logger.error(f"Undefined table attribute: [{name}:{domain}]")
            raise UndefinedTableAttribute
        return result.body[0]['Scalar']



"""
mmclass_nt.py - Generated named tuples corresponding to each MM class
"""


from collections import namedtuple


# Each tuple below defines the name of a class and its attributes that must
# be filled in to insert an instance of that class in the db.
#
# The _i part means "instance".
#
# The list of attributes of each tuple matches the corresponding .xcm file in
# the metamodel directory. Do not edit this file as it is automatically generated each
# time the metamodel is rebuilt.




# Computation subsystem
Computation_Action_i = namedtuple('Computation_Action_i', 'ID Activity Domain Operator')
Operand_i = namedtuple('Operand_i', 'Flow Activity Domain Computation')
Computed_Result_i = namedtuple('Computed_Result_i', 'Computation Activity Domain Flow')

# Class And Attribute subsystem
Alias_i = namedtuple('Alias_i', 'Name Class Domain')
Attribute_i = namedtuple('Attribute_i', 'Name Class Domain Scalar')
Attribute_In_Derivation_i = namedtuple('Attribute_In_Derivation_i', 'Source_attribute Source_class Target_attribute Target_class Domain')
Attribute_Reference_i = namedtuple('Attribute_Reference_i', 'From_attribute From_class To_attribute To_class Domain To_identifier Ref Rnum')
Class_i = namedtuple('Class_i', 'Name Cnum Domain')
Classifier_Attribute_i = namedtuple('Classifier_Attribute_i', 'Type Superclass Generalization Identifier Attribute Domain')
Derived_Attribute_i = namedtuple('Derived_Attribute_i', 'Name Class Domain')
Identifier_i = namedtuple('Identifier_i', 'Number Class Domain')
Identifier_Attribute_i = namedtuple('Identifier_Attribute_i', 'Identifier Attribute Class Domain')
Irreducible_Identifier_i = namedtuple('Irreducible_Identifier_i', 'Number Class Domain')
Non_Derived_Attribute_i = namedtuple('Non_Derived_Attribute_i', 'Name Class Domain')
Subclass_Classification_Value_i = namedtuple('Subclass_Classification_Value_i', 'Value Superclass Subclass Generalization Domain')
Super_Identifier_i = namedtuple('Super_Identifier_i', 'Number Class Domain')

# Type subsystem
Model_Attribute_i = namedtuple('Model_Attribute_i', 'Name Non_scalar_type Domain')
Scalar_i = namedtuple('Scalar_i', 'Name Domain')
Table_i = namedtuple('Table_i', 'Name Domain')
Table_Attribute_i = namedtuple('Table_Attribute_i', 'Name Table Domain Scalar')
Type_i = namedtuple('Type_i', 'Name Domain')

# Navigation subsystem
Association_Hop_i = namedtuple('Association_Hop_i', 'Number Path Domain')
Asymmetric_Circular_Hop_i = namedtuple('Asymmetric_Circular_Hop_i', 'Number Path Domain')
Circular_Hop_i = namedtuple('Circular_Hop_i', 'Number Path Domain Aggregation')
From_Asymmetric_Association_Class_Hop_i = namedtuple('From_Asymmetric_Association_Class_Hop_i', 'Number Path Domain')
Association_Class_Hop_i = namedtuple('Association_Class_Hop_i', 'Number Path Domain')
From_Symmetric_Association_Class_Hop_i = namedtuple('From_Symmetric_Association_Class_Hop_i', 'Number Path Domain')
Generalization_Hop_i = namedtuple('Generalization_Hop_i', 'Number Path Domain')
Hop_i = namedtuple('Hop_i', 'Number Path Domain Rnum Class_step')
Ordinal_Hop_i = namedtuple('Ordinal_Hop_i', 'Number Path Domain Ascending')
Path_i = namedtuple('Path_i', 'Name Domain Dest_class')
Perspective_Hop_i = namedtuple('Perspective_Hop_i', 'Number Path Domain Side Rnum')
Straight_Hop_i = namedtuple('Straight_Hop_i', 'Number Path Domain')
To_Subclass_Hop_i = namedtuple('To_Subclass_Hop_i', 'Number Path Domain')
To_Superclass_Hop_i = namedtuple('To_Superclass_Hop_i', 'Number Path Domain')
Symmetric_Hop_i = namedtuple('Symmetric_Hop_i', 'Number Path Domain')
To_Association_Class_Hop_i = namedtuple('To_Association_Class_Hop_i', 'Number Path Domain')
Traverse_Action_i = namedtuple('Traverse_Action_i', 'ID Activity Domain Path Source_flow Destination_flow')

# Action subsystem
Action_i = namedtuple('Action_i', 'ID Activity Domain Type')
Flow_Dependency_i = namedtuple('Flow_Dependency_i', 'From_action To_action Activity Domain Flow')
Wave_i = namedtuple('Wave_i', 'Number Activity Domain')
Wave_Assignment_i = namedtuple('Wave_Assignment_i', 'Action Activity Domain Wave')

# Relationship subsystem
Association_i = namedtuple('Association_i', 'Rnum Domain')
Association_Class_i = namedtuple('Association_Class_i', 'Rnum Class Domain Multiplicity')
Asymmetric_Perspective_i = namedtuple('Asymmetric_Perspective_i', 'Side Rnum Domain')
Binary_Association_i = namedtuple('Binary_Association_i', 'Rnum Domain')
Class_In_Lineage_i = namedtuple('Class_In_Lineage_i', 'Class Lnum Domain')
Constrained_Loop_i = namedtuple('Constrained_Loop_i', 'CLnum Domain')
Facet_i = namedtuple('Facet_i', 'Rnum Class Domain')
Generalization_i = namedtuple('Generalization_i', 'Rnum Domain Superclass')
Lineage_i = namedtuple('Lineage_i', 'Lnum Domain')
Loop_Segment_i = namedtuple('Loop_Segment_i', 'Loop Rnum Domain')
Minimal_Partition_i = namedtuple('Minimal_Partition_i', 'Rnum Domain A_subclass B_subclass')
Ordinal_Relationship_i = namedtuple('Ordinal_Relationship_i', 'Rnum Domain Ranked_class Ranking_attribute Ranking_identifier Ascending_perspective Descending_perspective')
P_Perspective_i = namedtuple('P_Perspective_i', 'Side Rnum Domain')
Perspective_i = namedtuple('Perspective_i', 'Side Rnum Domain Viewed_class Phrase Conditional Multiplicity')
Relationship_i = namedtuple('Relationship_i', 'Rnum Domain')
Subclass_i = namedtuple('Subclass_i', 'Rnum Class Domain')
Superclass_i = namedtuple('Superclass_i', 'Rnum Domain Class')
Symmetric_Perspective_i = namedtuple('Symmetric_Perspective_i', 'Side Rnum Domain')
T_Perspective_i = namedtuple('T_Perspective_i', 'Side Rnum Domain')
Unary_Association_i = namedtuple('Unary_Association_i', 'Rnum Domain')

# Domain subsystem
Domain_i = namedtuple('Domain_i', 'Name Alias')
Domain_Partition_i = namedtuple('Domain_Partition_i', 'Number Domain')
Element_i = namedtuple('Element_i', 'Label Domain')
Modeled_Domain_i = namedtuple('Modeled_Domain_i', 'Name')
Realized_Domain_i = namedtuple('Realized_Domain_i', 'Name')
Spanning_Element_i = namedtuple('Spanning_Element_i', 'Label Domain')
Subsystem_i = namedtuple('Subsystem_i', 'Name First_element_number Domain Alias')
Subsystem_Element_i = namedtuple('Subsystem_Element_i', 'Label Domain Subsystem')
System_i = namedtuple('System_i', 'Name')

# Select Action subsystem
Identifier_Select_i = namedtuple('Identifier_Select_i', 'ID Activity Domain Identifier Class')
Many_Select_i = namedtuple('Many_Select_i', 'ID Activity Domain Output_flow')
Select_Action_i = namedtuple('Select_Action_i', 'ID Activity Domain Input_flow')
Single_Select_i = namedtuple('Single_Select_i', 'ID Activity Domain Output_flow')
Zero_One_Cardinality_Select_i = namedtuple('Zero_One_Cardinality_Select_i', 'ID Activity Domain')

# Event subsystem
Delegated_Event_i = namedtuple('Delegated_Event_i', 'Name Class Domain')
Delegation_On_Generalization_i = namedtuple('Delegation_On_Generalization_i', 'Delegated_event Superclass Domain Generalization')
Effective_Event_i = namedtuple('Effective_Event_i', 'Name State_model Domain')
Event_i = namedtuple('Event_i', 'Name State_model Domain')
Event_Specification_i = namedtuple('Event_Specification_i', 'Name State_model Domain State_signature')
Forwarded_Delegated_Event_i = namedtuple('Forwarded_Delegated_Event_i', 'Name Class Domain')
Forwarded_Effective_Event_i = namedtuple('Forwarded_Effective_Event_i', 'Name Class Domain')
Forwarded_Event_i = namedtuple('Forwarded_Event_i', 'Name Subclass Domain Superclass Generalization')
Monomorphic_Event_i = namedtuple('Monomorphic_Event_i', 'Name State_model Domain')
Monomorphic_Event_Specification_i = namedtuple('Monomorphic_Event_Specification_i', 'Name State_model Domain')
Polymorphic_Event_Specification_i = namedtuple('Polymorphic_Event_Specification_i', 'Name Class Domain')
Root_Delegated_Event_i = namedtuple('Root_Delegated_Event_i', 'Name Class Domain')

# Flow subsystem
Class_Accessor_i = namedtuple('Class_Accessor_i', 'Class Activity Domain Output_flow')
Control_Flow_i = namedtuple('Control_Flow_i', 'ID Activity Domain')
Data_Flow_i = namedtuple('Data_Flow_i', 'ID Activity Domain')
Flow_i = namedtuple('Flow_i', 'ID Activity Domain')
Gate_Action_i = namedtuple('Gate_Action_i', 'ID Activity Domain Output_flow')
Gate_Input_i = namedtuple('Gate_Input_i', 'Gate_action Input_flow Activity Domain')
Instance_Flow_i = namedtuple('Instance_Flow_i', 'ID Activity Domain Class')
Labeled_Flow_i = namedtuple('Labeled_Flow_i', 'ID Activity Domain Name')
Multiple_Instance_Flow_i = namedtuple('Multiple_Instance_Flow_i', 'ID Activity Domain')
Non_Scalar_Flow_i = namedtuple('Non_Scalar_Flow_i', 'ID Activity Domain')
Table_Flow_i = namedtuple('Table_Flow_i', 'ID Activity Domain')
Scalar_Flow_i = namedtuple('Scalar_Flow_i', 'ID Activity Domain Type')
Single_Instance_Flow_i = namedtuple('Single_Instance_Flow_i', 'ID Activity Domain')
Relation_Flow_i = namedtuple('Relation_Flow_i', 'ID Activity Domain Type')
Tuple_Flow_i = namedtuple('Tuple_Flow_i', 'ID Activity Domain')
Unlabeled_Flow_i = namedtuple('Unlabeled_Flow_i', 'ID Activity Domain')

# Attribute Access subsystem
Attribute_Read_Access_i = namedtuple('Attribute_Read_Access_i', 'Attribute Class Read_action Activity Domain Output_flow')
Attribute_Write_Access_i = namedtuple('Attribute_Write_Access_i', 'Attribute Class Write_action Activity Domain Input_flow')
Read_Action_i = namedtuple('Read_Action_i', 'ID Activity Domain Instance_flow')
Write_Action_i = namedtuple('Write_Action_i', 'ID Activity Domain Instance_flow')

# Formalization subsystem
Association_Class_Reference_i = namedtuple('Association_Class_Reference_i', 'Ref_type Association_class Participating_class Rnum Domain')
Association_Reference_i = namedtuple('Association_Reference_i', 'Ref_type From_class To_class Rnum Domain Perspective')
Formalizing_Class_Role_i = namedtuple('Formalizing_Class_Role_i', 'Rnum Class Domain')
Generalization_Reference_i = namedtuple('Generalization_Reference_i', 'Subclass Superclass Rnum Domain Ref_type')
Referring_Class_i = namedtuple('Referring_Class_i', 'Rnum Class Domain')
Reference_i = namedtuple('Reference_i', 'Ref From_class To_class Rnum Domain')
Simple_Association_Reference_i = namedtuple('Simple_Association_Reference_i', 'From_class To_class Rnum Domain Ref_type')
P_Reference_i = namedtuple('P_Reference_i', 'Association_class Participating_class Rnum Domain Ref_type')
T_Reference_i = namedtuple('T_Reference_i', 'Association_class Participating_class Rnum Domain Ref_type')

# Conditional Logic subsystem
Case_i = namedtuple('Case_i', 'Flow Activity Domain Switch_action')
Control_Dependency_i = namedtuple('Control_Dependency_i', 'Control_flow Action Activity Domain')
Decision_Action_i = namedtuple('Decision_Action_i', 'ID Activity Domain Predicate')
Decision_Input_i = namedtuple('Decision_Input_i', 'Decision_action Flow Activity Domain')
Emitter_i = namedtuple('Emitter_i', 'Output_flow Activity Domain')
Emitter_Input_i = namedtuple('Emitter_Input_i', 'Input_flow Output_flow Activity Domain')
Match_Value_i = namedtuple('Match_Value_i', 'Case_flow Activity Domain Value')
Result_i = namedtuple('Result_i', 'Decision Decision_action Flow Activity Domain')
Scalar_Switch_Action_i = namedtuple('Scalar_Switch_Action_i', 'ID Activity Domain Scalar_input')
Sequence_Flow_i = namedtuple('Sequence_Flow_i', 'Source_action Flow Activity Domain')
Subclass_Switch_Action_i = namedtuple('Subclass_Switch_Action_i', 'ID Activity Domain Superclass_instance Generalization')
Switch_Action_i = namedtuple('Switch_Action_i', 'ID Activity Domain')

# Activity subsystem
Activity_i = namedtuple('Activity_i', 'Anum Domain')
Asynchronous_Activity_i = namedtuple('Asynchronous_Activity_i', 'Anum Domain')
Method_i = namedtuple('Method_i', 'Anum Name Class Domain Executing_instance_flow')
Method_Signature_i = namedtuple('Method_Signature_i', 'SIGnum Method Class Domain')
Operation_Signature_i = namedtuple('Operation_Signature_i', 'SIGnum EE Operation Domain')
Parameter_i = namedtuple('Parameter_i', 'Name Signature Domain Input_flow Activity Type')
Signature_i = namedtuple('Signature_i', 'SIGnum Domain')
State_Activity_i = namedtuple('State_Activity_i', 'Anum State State_model Domain')
Synchronous_Activity_i = namedtuple('Synchronous_Activity_i', 'Anum Domain')
Synchronous_Output_i = namedtuple('Synchronous_Output_i', 'Anum Domain Output_flow Type')

# Relational Action subsystem
Class_Restriction_Condition_i = namedtuple('Class_Restriction_Condition_i', 'Select_action Activity Domain')
Criterion_i = namedtuple('Criterion_i', 'ID Action Activity Domain Attribute Non_scalar_type')
Comparison_Criterion_i = namedtuple('Comparison_Criterion_i', 'ID Action Activity Attribute Domain Value Comparison')
Equivalence_Criterion_i = namedtuple('Equivalence_Criterion_i', 'ID Action Activity Attribute Domain Operation Value Scalar')
Extend_Action_i = namedtuple('Extend_Action_i', 'ID Activity Domain Added_attribute Table Result')
Extract_Action_i = namedtuple('Extract_Action_i', 'ID Activity Domain Input_tuple Table Attribute Output_scalar')
Project_Action_i = namedtuple('Project_Action_i', 'ID Activity Domain')
Projected_Attribute_i = namedtuple('Projected_Attribute_i', 'Attribute Non_scalar_type Project_action Activity Domain')
Rank_Restrict_Action_i = namedtuple('Rank_Restrict_Action_i', 'ID Activity Domain Attribute Non_scalar_type Selection_cardinality Extent')
Relational_Action_i = namedtuple('Relational_Action_i', 'ID Activity Domain')
Rename_Action_i = namedtuple('Rename_Action_i', 'ID Activity Domain From_attribute From_non_scalar_type To_attribute To_table')
Restrict_Action_i = namedtuple('Restrict_Action_i', 'ID Activity Domain')
Restriction_Condition_i = namedtuple('Restriction_Condition_i', 'Action Activity Domain Expression Selection_cardinality')
Set_Action_i = namedtuple('Set_Action_i', 'ID Operation Activity Domain Input_b_flow')
Source_Tuple_Attribute_Access_i = namedtuple('Source_Tuple_Attribute_Access_i', 'Extend_action Model_attribute Non_scalar_type Activity Domain Flow')
Table_Action_i = namedtuple('Table_Action_i', 'ID Activity Domain Input_a_flow Output_flow')
Table_Restriction_Condition_i = namedtuple('Table_Restriction_Condition_i', 'Restrict_action Activity Domain')

# External Entity subsystem
Asynchronous_Operation_i = namedtuple('Asynchronous_Operation_i', 'Name EE Domain Anum')
External_Entity_i = namedtuple('External_Entity_i', 'EEnum Name Class Domain')
Operation_i = namedtuple('Operation_i', 'Name EE Domain Direction')
Synchronous_Operation_i = namedtuple('Synchronous_Operation_i', 'Name EE Domain Anum')

# State subsystem
Assigner_i = namedtuple('Assigner_i', 'Rnum Domain')
Deletion_State_i = namedtuple('Deletion_State_i', 'Name Class Domain')
Event_Response_i = namedtuple('Event_Response_i', 'State Event State_model Domain')
Initial_Pseudo_State_i = namedtuple('Initial_Pseudo_State_i', 'Name Class Domain')
Initial_Transition_i = namedtuple('Initial_Transition_i', 'From_state Class Domain Event')
Lifecycle_i = namedtuple('Lifecycle_i', 'Class Domain')
Multiple_Assigner_i = namedtuple('Multiple_Assigner_i', 'Rnum Domain Partitioning_class')
Non_Deletion_State_i = namedtuple('Non_Deletion_State_i', 'Name State_model Domain')
Non_Transition_i = namedtuple('Non_Transition_i', 'State Event State_model Domain Behavior Reason')
Real_State_i = namedtuple('Real_State_i', 'Name State_model Domain Signature Activity')
Single_Assigner_i = namedtuple('Single_Assigner_i', 'Rnum Domain')
State_i = namedtuple('State_i', 'Name State_model Domain')
State_Model_i = namedtuple('State_Model_i', 'Name Domain')
State_Signature_i = namedtuple('State_Signature_i', 'SIGnum State_model Domain')
Transition_i = namedtuple('Transition_i', 'From_state Event State_model Domain To_state')


"""
ordinal.py – Process a parsed ordinal relationship to populate the metamodel db
"""

# System
import logging

# Model Integration
from pyral.relvar import Relvar

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.mmclass_nt import Ordinal_Relationship_i

_logger = logging.getLogger(__name__)

class Ordinal:
    """
    Populate all relevant Ordinal Relationship relvars
    """
    rnum = None
    ascend = None
    oform = None

    def __init__(self, tr: str, domain: str, rnum: str, record):
        """
        Populate an Ordinal Relationship

        :param tr: The name of the open transaction
        :param domain: The domain name
        :param rnum: The relationship name
        :param record: The relationship parse record
        """

        self.rnum = rnum
        self.ascend = record['ascend']
        self.oform = record['oform']

        # Populate
        _logger.info(f"Populating Ordinal [{self.rnum}]")
        Relvar.insert(db=mmdb, tr=tr, relvar='Ordinal_Relationship', tuples=[
            Ordinal_Relationship_i(Rnum=self.rnum, Domain=domain, Ranked_class=self.ascend['cname'],
                                   Ranking_attribute=self.oform['ranking attr']['name'], Ranking_identifier=self.oform['id'],
                                   Ascending_perspective=self.ascend['highval'],
                                   Descending_perspective=self.ascend['lowval']
                                   )
        ])



"""
action.py – Populate an action superclass instance in PyRAL
"""

import logging
from typing import Set, Dict, List, Optional
from xuml_populate.config import mmdb
from scrall.parse.visitor import PATH_a
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction
from xuml_populate.populate.mmclass_nt import Action_i, Wave_i
from collections import namedtuple


_logger = logging.getLogger(__name__)

class Action:
    """
    A metamodel action
    """

    next_action_id = {}

    @classmethod
    def populate(cls, tr: str, anum: str, domain: str, action_type: str) -> str:
        """

        :param tr: Open db transaction
        :param anum: The Activity Number
        :param domain: Name of the domain
        :param action_type:  The subclass of this action
        :return: The action id
        """
        activity_key = f'{domain}:{anum}'  # combine attributes to get id
        if activity_key not in cls.next_action_id.keys():
            # Initialize the Action ID counter for a new Activity (anum)
            cls.next_action_id[activity_key] = 0
        # Get the next action ID for this Activity
        cls.next_action_id[activity_key] += 1
        actn_id = f'ACTN{cls.next_action_id[activity_key]}'

        # Now populate an instance of Action
        _logger.info("Transaction open: Action")


        # Populate the Statement superclass
        Relvar.insert(db=mmdb, tr=tr, relvar='Action', tuples=[
            Action_i(ID=actn_id, Activity=anum, Domain=domain, Type=action_type)
        ])
        return actn_id



"""
class_exceptions.py – Exceptions encountered when loading a model
"""

# Every error should have the same format
# with a standard prefix and postfix defined here
pre = "\nModel loader -- "
post = " --"

# ---
# ---

class UserModel(Exception):
    pass

class ReferenceToNonIdentifier(UserModel):
    def __str__(self):
        return f"{pre}Attribute does not reference an identifier in target"

class MixedTargetID(UserModel):
    def __str__(self):
        return f"{pre}Attributes from more than one ID in target reference"



"""
action_exceptions.py – Errors that occur while populating actions
"""

# Every error should have the same format
# with a standard prefix and postfix defined here
pre = "\nScrall populator: -- "
post = " --"


class ActionException(Exception):
    pass

class UndefinedMethod(ActionException):
    pass

class UndefinedSynchOutputExecutionUnit(ActionException):
    pass

class MethodXIFlowNotPopulated(ActionException):
    pass

class UndefinedAttribute(ActionException):
    pass

class UndefinedParameter(ActionException):
    pass
class UndefinedTableAttribute(ActionException):
    pass

class SwitchException(ActionException):
    pass

class BadScalarSwitchInput(SwitchException):
    pass

class FlowException(ActionException):
    pass

class ControlFlowHasNoTargetActions(FlowException):
    pass

class NonScalarFlowRequired(FlowException):
    pass

class TableExpressionException(ActionException):
    pass

class ScalarExpressionException(ActionException):
    pass

class ScalarOperationOrExpressionExpected(ScalarExpressionException):
    pass

class ScalarAssignmentfromMultipleTuples(ScalarExpressionException):
    pass
class ScalarAssignmentFlowMismatch(ScalarExpressionException):
    pass

class UndefinedHeaderExpressionOp(TableExpressionException):
    pass

class ProductForbidsCommonAttributes(TableExpressionException):
    pass

class SetOpRequiresSameHeaders(TableExpressionException):
    pass

class UnjoinableHeaders(TableExpressionException):
    pass

class TraversalActionException(ActionException):
    pass

class SelectActionException(ActionException):
    pass

class NoInputInstanceFlow(SelectActionException):
    pass

class ProjectedAttributeNotDefined(TableExpressionException):
    pass

class TableOperationOrExpressionExpected(TableExpressionException):
    pass

class SelectionOnScalarFlow(ActionException):
    def __init__(self, path, text, x):
        self.path = path
        self.text = text[slice(*x)]

    def __str__(self):
        return f'{pre}Cannot select/restrict on a scalar flow. Verify input to selection phrase' \
               f'{post}\n\t[{self.path}] >> {self.text}'

class AssignZeroOneInstanceHasMultiple(ActionException):
    def __init__(self, path, text, x):
        self.path = path
        self.text = text[slice(*x)]

    def __str__(self):
        return f'{pre}Single instance assignment operator may recieve multiple instances, try ..= operator or ' \
               f'modify RHS to yield at most one instance{post}\n\t[{self.path}] >> {self.text}'

class NoClassOrInstanceFlowForInstanceSetName(ActionException):
    def __init__(self, path, text, x):
        self.path = path
        self.text = text[slice(*x)]

    def __str__(self):
        return f'{pre}Name in instance set is neither a class or an instance flow source' \
               f'{post}\n\t[{self.path}] >> {self.text}'

class ComparingNonAttributeInSelection(SelectActionException):
    pass

class NoDestinationInPath(ActionException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f'{pre}No destination class at end of path: [{self.path}].{post}'

class UndefinedClass(ActionException):
    def __init__(self, cname):
        self.cname = cname

    def __str__(self):
        return f'{pre}Class [{self.cname}] not defined.{post}'

class IncompletePath(ActionException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f'{pre}Path requires at least one hop: [{self.path}].{post}'

class NoPathFromClass(ActionException):
    def __init__(self, rnum, domain):
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Undefined relationship [{self.rnum}] in domain [{self.domain}].{post}'

class UndefinedRelationship(ActionException):
    def __init__(self, rnum, domain):
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Undefined relationship [{self.rnum}] in domain [{self.domain}].{post}'

class UndefinedAssociation(ActionException):
    def __init__(self, rnum, domain):
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Undefined association [{self.rnum}] in domain [{self.domain}].{post}'

class PerspectiveNotDefined(TraversalActionException):
    def __init__(self, phrase,domain):
        self.phrase = phrase
        self.domain = domain

    def __str__(self):
        return f'{pre}Perspective [{self.phrase}] not defined in domain [{self.domain}].{post}'

class RelationshipUnreachableFromClass(TraversalActionException):
    def __init__(self, rnum, cname, domain):
        self.cname = cname
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Unreachable relationship [{self.rnum}] from [{self.cname}] in domain [{self.domain}].{post}'

class NeedPerspectiveToHop(TraversalActionException):
    def __init__(self, rnum, domain):
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Reflexive association [{self.rnum}] in domain [{self.domain}] cannot be resolved' \
               f'without perspective.{post}'
class NeedPerspectiveOrClassToHop(TraversalActionException):
    def __init__(self, rnum, domain):
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Association [{self.rnum}] in domain [{self.domain}] cannot be resolved' \
               f'without perspective or class.{post}'

class HopToUnreachableClass(TraversalActionException):
    def __init__(self, cname, rnum, domain):
        self.cname = cname
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Relationship [{self.rnum}] does not reach class [{self.cname}] in domain [{self.domain}].{post}'

class SubclassNotInGeneralization(TraversalActionException):
    def __init__(self, subclass, rnum, domain):
        self.subclass = subclass
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Generalization [{self.rnum}] does not include subclass [{self.subclass}] in domain' \
               f'[{self.domain}].{post}'
class NoSubclassInHop(TraversalActionException):
    def __init__(self, superclass, rnum, domain):
        self.superclass = superclass
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}Generalization [{self.rnum}] from [{self.superclass}] does not reach a subclass in domain' \
               f'[{self.domain}].{post}'
class MissingTorPrefInAssociativeRel(TraversalActionException):
    def __init__(self, rnum, domain):
        self.rnum = rnum
        self.domain = domain

    def __str__(self):
        return f'{pre}P or T ref not found for associative relationship [{self.rnum}] in domain [{self.domain}].{post}'

class UnexpectedClassOrPerspectiveInPath(TraversalActionException):
    def __init__(self, name, path):
        self.name = name
        self.path = path

    def __str__(self):
        return f'{pre}Name [{self.name}] not expected in path [{self.path}].{post}'




""" sexpr.py -- Walk through a scalar expression and populate elements """

import logging
from typing import List
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.expressions.instance_set import InstanceSet
from xuml_populate.populate.actions.read_action import ReadAction
from xuml_populate.exceptions.action_exceptions import ScalarOperationOrExpressionExpected
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap, Boundary_Actions
from scrall.parse.visitor import Scalar_RHS_a, MATH_a, BOOL_a, INST_a, N_a, Projection_a, Op_chain_a, INST_PROJ_a
from xuml_populate.populate.actions.project_action import ProjectAction

from pyral.relation import Relation  # Keep for debugging

_logger = logging.getLogger(__name__)


class ScalarExpr:
    """
    For reference, a scalar expression in the scrall grammar consists of terms

    term = (NOT SP*)? UNARY_MINUS? (scalar / "(" SP* sexpr SP* ")")
    scalar = value / QTY? scalar_chain
    scalar_chain = (ITS op_chain) / ((scalar_source / instance_set projection?) op_chain?)
    scalar_source = type_selector / input_param
    op_chain = ('.' (scalar_op / name))*
    scalar_op = name supplied_params

    So we need to walk through the parse tree through the nested operations, possibly
    building instance sets.
    """
    text = None  # A text representation of the expression
    activity_data = None
    component_flow = None
    output_tflow_id = None
    action_inputs = None
    action_outputs = None
    anum = None
    domain = None

    @classmethod
    def process(cls, rhs: Scalar_RHS_a, input_instance_flow: Flow_ap,
                activity_data: Activity_ap) -> (Boundary_Actions, List[Flow_ap]):
        """
        Walks through a scalar expression on the right hand side of a scalar assignment to
        obtain a tuple flow with one or more attributes. Each attribute value will be assigned.
        The order in which attributes are specified in the action language is returned along with
        the tuple flow.

        :param rhs: The right hand side of a table assignment
        :param input_instance_flow:
        :param activity_data:
        :return:  The output tuple flow and the attribute names as ordered in the RHS text expression
        """
        cls.domain = activity_data.domain
        cls.anum = activity_data.anum
        cls.activity_data = activity_data

        cls.action_outputs = {}  # ID's of all Action output Data Flows
        cls.action_inputs = {}  # ID's of all Action input Data Flows

        rhs = cls.walk(sexpr=rhs.expr, input_flow=input_instance_flow)

        all_ins = {v for s in cls.action_inputs.values() for v in s}
        all_outs = {v for s in cls.action_outputs.values() for v in s}
        init_aids = {a for a in cls.action_inputs.keys() if not cls.action_inputs[a].intersection(all_outs)}
        final_aids = {a for a in cls.action_outputs.keys() if not cls.action_outputs[a].intersection(all_ins)}

        return Boundary_Actions(ain=init_aids, aout=final_aids), rhs

    @classmethod
    def resolve_iset(cls, iset: INST_a, op_chain: Op_chain_a = None, projection: Projection_a = None) -> List[Flow_ap]:
        pass

    @classmethod
    def walk(cls, sexpr: INST_PROJ_a | MATH_a | BOOL_a | N_a, input_flow: Flow_ap) -> [Flow_ap]:
        """

        :param sexpr:  Parsed scalar expression
        :param input_flow:
        :return:  Output scalar flow
        """
        component_flow = input_flow
        match type(sexpr).__name__:
            case 'INST_PROJ_a':
                action_input = component_flow
                initial_aid, final_aid, component_flow = InstanceSet.process(input_instance_flow=action_input,
                                                                             iset_components=sexpr.iset.components,
                                                                             activity_data=cls.activity_data)
                # Add the output flow generated by the instance set expression to the set of ouput flows
                if initial_aid:
                    # For an InstanceSet with a single labled flow component, no action is created
                    # So don't process action inputs and outputs unless there is an initial_aid
                    cls.action_inputs[initial_aid] = {action_input.fid}
                    if final_aid:
                        cls.action_outputs[final_aid] = {component_flow.fid}
                action_input = component_flow
                project_attrs = tuple([a.name for a in sexpr.projection.attrs])
                aid, sflows = ReadAction.populate(input_single_instance_flow=action_input,
                                                  attrs=project_attrs, anum=cls.anum, domain=cls.domain)
                cls.action_inputs[aid] = {action_input.fid}
                cls.action_outputs[aid] = {s.fid for s in sflows}
                return sflows
            case 'N_a':
                pass
            case 'BOOL_a':
                pass
            case 'MATH_a':
                action_input = component_flow
                operand_flows = []
                op_name = sexpr.op
                for o in sexpr.operands:
                    match type(o).__name__:
                        case 'INST_PROJ_a':
                            component_flow = InstanceSet.process(input_instance_flow=action_input,
                                                                 iset_components=o.iset.components,
                                                                 activity_data=cls.activity_data)
                            if o.iset.select:
                                pass
                            if o.projection:
                                tflow = ProjectAction.populate(projection=o.projection,
                                                               input_nsflow=action_input,
                                                               activity_data=cls.activity_data)
                                pass
                            pass
                        case _:
                            pass

                    operand_flows.append(cls.walk(sexpr=o, input_flow=component_flow))
                pass
            case 'Op_chain_a':
                pass
            case _:
                _logger.error(
                    f"Expected .... but received {type(sexpr).__name__} during sexpr walk")
                raise ScalarOperationOrExpressionExpected
        # Process optional header, selection, and projection actions for the TEXPR
        return component_flow



"""
table_assignment.py – Populate elements of a table assignment
"""

import logging
from typing import Dict, Set
from xuml_populate.config import mmdb
from xuml_populate.populate.mmclass_nt import Labeled_Flow_i
from xuml_populate.populate.actions.expressions.table_expr import TableExpr
from xuml_populate.populate.actions.aparse_types import (Flow_ap, MaxMult, Content, Activity_ap, Boundary_Actions,
                                                         Labeled_Flow)
from scrall.parse.visitor import Table_Assignment_a

from pyral.relvar import Relvar
from pyral.relation import Relation  # Keep for debugging
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Migrate = "Migrate to Label"


class TableAssignment:
    """
    Break down a table assignment statement into action semantics and populate them

    """

    input_instance_flow = None  # The instance flow feeding the next component on the RHS
    input_instance_ctype = None  # The class type of the input instance flow
    domain = None
    anum = None
    mmdb = None
    activity_path = None
    scrall_text = None

    @classmethod
    def process(cls, activity_data: Activity_ap, table_assign_parse: Table_Assignment_a,
                case_name: str, case_outputs: Set[Labeled_Flow] = None) -> Boundary_Actions:
        """
        Given a parsed table assignment consisting of an LHS and an RHS, populate each component action
        and return the resultant table flow

        We'll need an initial flow and we'll need to create intermediate instance flows to connect the components.
        The final output flow must be a table flow. The associated Table determines the type of the
        assignment. If the LHS spcifies an explicit Table, the resultant Table which must match.

        :param activity_data:
        :param table_assign_parse: A parsed table assignment
        :param case_name:
        :param case_outputs:
        """

        lhs = table_assign_parse.lhs
        rhs = table_assign_parse.rhs
        cls.input_instance_flow = activity_data.xiflow

        # The executing instance is by nature a single instance flow
        xi_flow = Flow_ap(fid=activity_data.xiflow, content=Content.INSTANCE, tname=activity_data.cname,
                          max_mult=MaxMult.ONE)

        bactions, output_flow = TableExpr.process(tuple_output=table_assign_parse.assign_tuple, rhs=rhs, activity_data=activity_data, input_instance_flow=xi_flow)

        case_prefix = '' if not case_name else f"{case_name}_"
        output_flow_label = case_prefix + lhs
        if case_name:
            case_outputs.add(Labeled_Flow(label=lhs, flow=output_flow))
        # TODO: handle case where lhs is an explicit table assignment

        # Migrate the output_flow to a labeled flow
        _logger.info(f"Labeling output of table expression to [{lhs}]")
        Transaction.open(db=mmdb, name=tr_Migrate)
        # Delete the Unlabeled flow
        Relvar.deleteone(db=mmdb, tr=tr_Migrate, relvar_name="Unlabeled_Flow",
                         tid={"ID": output_flow.fid, "Activity": activity_data.anum, "Domain": activity_data.domain})
        # Insert the labeled flow
        Relvar.insert(db=mmdb, tr=tr_Migrate, relvar='Labeled_Flow', tuples=[
            Labeled_Flow_i(ID=output_flow.fid, Activity=activity_data.anum, Domain=activity_data.domain,
                           Name=output_flow_label)
        ])
        Transaction.execute(db=mmdb, name=tr_Migrate)
        return bactions



""" table.py - Define a table """

import logging
from typing import Tuple, Dict
from xuml_populate.config import mmdb
from pyral.transaction import Transaction
from pyral.relvar import Relvar
from pyral.relation import Relation
from xuml_populate.populate.mmclass_nt import Table_i, Type_i, Table_Attribute_i, Model_Attribute_i

_logger = logging.getLogger(__name__)

class Table:
    """
    Populates the Table and all unconditionally related classes.
    This includes Table Attribute, Model Attribute and Table Flow
    """

    @classmethod
    def header(cls, tname: str, domain: str) -> Dict[str, str]:
        """
        Returns the header for this table type

        :param tname:   Class name
        :param domain:  Domain name
        :return:  Dictionary of attr:scalar (type) values
        """
        R = f"Table:<{tname}>, Domain:<{domain}>"
        attrs = Relation.restrict(mmdb, relation='Table_Attribute', restriction=R)
        h = {a['Name']: a['Scalar'] for a in attrs.body}
        return h

    @classmethod
    def populate(cls, tr: str, table_header: Dict[str, str], domain: str) -> str:
        """
        Population happens in the context of a Relation Flow transaction.

        :param tr: The Relaton Flow transaction name
        :param table_header: A dictionary of attribute name;type pairs
        :param domain: The domain name
        :return: The name of the Table
        """
        # Generate a table name by converting the table_header into one long string
        # with attribute name/types delimited by underscores
        table_name = "_".join([f"{attr_name}_{attr_type}" for attr_name, attr_type in table_header.items()])

        # Check to see if the table already exists, if so, just return the name
        R = f"Name:<{table_name}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Table', restriction=R)
        if result.body:
            return table_name

        _logger.info(f"Populating Table flow on existing Table: [{table_name}]")
        # A Table can't exist without a flow
        Transaction.open(mmdb, tr)  # Table type
        Relvar.insert(mmdb, tr=tr, relvar='Table', tuples=[
            Table_i(table_name, domain)
        ])
        Relvar.insert(mmdb, tr=tr, relvar='Type', tuples=[
            Type_i(table_name, domain)
        ])
        for a, t in table_header.items():
            Relvar.insert(mmdb, tr=tr, relvar='Table_Attribute', tuples=[
                Table_Attribute_i(Name=a, Table=table_name, Domain=domain, Scalar=t)
            ])
            Relvar.insert(mmdb, tr=tr, relvar='Model_Attribute', tuples=[
                Model_Attribute_i(Name=a, Non_scalar_type=table_name, Domain=domain)
            ])
        return table_name




""" signature.py – Populate a Signature in the metamodel db """

import logging
from pyral.relvar import Relvar
from xuml_populate.config import mmdb
from xuml_populate.populate.element import Element
from xuml_populate.populate.mmclass_nt import Signature_i

_logger = logging.getLogger(__name__)

class Signature:
    """
    Populate a Signature relvar with a unique id
    """
    @classmethod
    def populate(cls, tr: str, subsys: str, domain: str) -> str:
        """

        :param tr: The name of the open transaction
        :param subsys: The subsystem name
        :param domain: The domain name
        :return: The Signature id (SIGnum)
        """

        # Populate
        SIGnum = Element.populate_unlabeled_subsys_element(tr=tr, prefix='SIG', subsystem=subsys, domain=domain)
        Relvar.insert(mmdb, tr=tr, relvar='Signature', tuples=[
            Signature_i(SIGnum=SIGnum, Domain=domain)
        ])
        return SIGnum

        # Just enough here to support Activity Subsystem population
        # TODO: Add the rest of this subsystem later


"""
class_accessor.py – Populate a Class Accessor action instance in PyRAL
"""

import logging
from xuml_populate.config import mmdb
from typing import Optional
from xuml_populate.populate.mmclass_nt import Class_Accessor_i
from xuml_populate.populate.flow import Flow
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

tr_Class_Accessor = "Class Accessor"

class ClassAccessor:
    """
    Create all relations for a Class Accessor
    """

    @classmethod
    def populate(cls, name: str, anum: str, domain: str) -> Optional[str]:
        """
        Populate the Class Accessor Action or return nothing if the name is not a class

        :param name:  A name which may or may not correspond to a class
        :param anum:  The enclosing anum
        :param domain:  The current domain
        :return: The output flow id of an existing or newly populated class accessor or none name is not a class
        """
        # Return None if the name does not match any defined Class
        R = f"Name:<{name}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Class', restriction=R)
        if not result.body:
            return None

        # Return flow id of existing Class Accessor
        R = f"Class:<{name}>, Activity:<{anum}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Class_Accessor', restriction=R)
        if result.body:
            return result.body[0].fid

        # Populate a Class Accessor and Multiple Instance Flow returning the flow id
        Transaction.open(mmdb, tr_Class_Accessor)
        output_flow = Flow.populate_instance_flow(cname=name, anum=anum, domain=domain, label=None)
        Relvar.insert(mmdb, relvar='Class_Accessor', tuples=[
            Class_Accessor_i(Class=name, Activity=anum, Domain=domain, Output_flow=output_flow.fid)
        ])
        Transaction.execute(mmdb, tr_Class_Accessor)
        return output_flow.fid



"""
switch_statement.py – Populate a switch action instance in PyRAL
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.exceptions.action_exceptions import ActionException, BadScalarSwitchInput
from xuml_populate.populate.actions.aparse_types import Activity_ap, Boundary_Actions
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.flow import Flow
from scrall.parse.visitor import Switch_a
from xuml_populate.populate.mmclass_nt import (Switch_Action_i, Scalar_Switch_Action_i, Case_i, Match_Value_i,
                                               Control_Dependency_i, Result_i, Sequence_Flow_i,
                                               Decision_Input_i, Decision_Action_i,
                                               Subclass_Switch_Action_i, Gate_Action_i, Gate_Input_i)
from pyral.relvar import Relvar
from pyral.relation import Relation  # Keep here for debugging
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

from collections import namedtuple, Counter

Case_Control = namedtuple("Case_Control", "match_values target_actions")

# Transactions
tr_Switch = "Switch Action"

class SwitchStatement:
    """
    Create all relations for a Switch Action
    """

    output_actions = None
    labeled_outputs = None

    @classmethod
    def populate(cls, sw_parse: Switch_a, activity_data: Activity_ap) -> Boundary_Actions:
        """
        Populate the Switch Action and its Cases

        :param sw_parse:  The parsed switch action group
        :param activity_data: For descriptive error messages
        """
        anum = activity_data.anum
        domain = activity_data.domain
        cls.output_actions = set()
        cls.labeled_outputs = {}  # { <case>: [output_flow, ...], }  labeled output flows keyed by case

        # Process the input flow
        scalar_input_flow = None
        match type(sw_parse.input_flow).__name__:
            case 'IN_a':
                # Verify that this is a scalar flow and get its flow id based on the label
                scalar_input_flow = Flow.find_labeled_scalar_flow(name=sw_parse.input_flow.name, anum=anum,
                                                                  domain=domain)
                if not scalar_input_flow:
                    _logger.error(f"Scalar switch on parameter [{sw_parse.input_flow.name}] is not a scalar flow")
                    raise BadScalarSwitchInput
            case 'INST_PROJ_a':
                pass
            case _:
                # TODO: implement other cases (N_a, ?)
                pass
        # For each case, create actions (open/close transaction) and create list of action ids
        cactions = {}
        for c in sw_parse.cases:
            # Need to create a Control Flow / Case instnace per case_value and label it with the case_value name
            # Combined with the action_id to ensure label is unique within the Activity
            # Need to create all component set statements and obtain a set of initial action ids
            if c.comp_statement_set.statement:
                case_name = f"{'_'.join(c.enums)}"
                cls.labeled_outputs[case_name] = set()
                from xuml_populate.populate.statement import Statement
                boundary_actions = Statement.populate(activity_data=activity_data,
                                                      statement_parse=c.comp_statement_set.statement,
                                                      case_name=case_name,
                                                      case_outputs=cls.labeled_outputs[case_name])
                cls.output_actions = cls.output_actions.union(boundary_actions.aout)
                cactions[case_name] = Case_Control(match_values=c.enums, target_actions=boundary_actions.ain)
            else:
                if not c.component_statement_set.block:
                    raise ActionException
                # process block
                pass
            pass
        # Populate the Action superclass instance and obtain its action_id
        Transaction.open(db=mmdb, name=tr_Switch)
        action_id = Action.populate(tr=tr_Switch, anum=anum, domain=domain, action_type="scalar switch")  # Transaction open
        # TODO: It appears that the Subclass Switch Action has is not yet populated/implemented
        Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Switch_Action', tuples=[
            Switch_Action_i(ID=action_id, Activity=anum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Scalar_Switch_Action', tuples=[
            Scalar_Switch_Action_i(ID=action_id, Activity=anum, Domain=domain,
                                   Scalar_input=scalar_input_flow.fid)
        ])
        for k, v in cactions.items():
            control_flow_fid = Flow.populate_control_flow(tr=tr_Switch, label=k, enabled_actions=v.target_actions,
                                                          anum=anum, domain=domain)
            Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Case', tuples=[
                Case_i(Flow=control_flow_fid, Activity=anum, Domain=domain, Switch_action=action_id)
            ])
            for mv in v.match_values:
                Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Match_Value', tuples=[
                    Match_Value_i(Case_flow=control_flow_fid, Activity=anum, Domain=domain, Value=mv)
                ])
                pass
            # TODO: Create cases and control dependencies

        # Get all labels that appear in more than one case
        # (The same label cannot appear more than once in the same case since label names are unique within a case)
        # First create a bag of labels (list with possible duplicates) ranging over all cases
        all_labels = [lf.label for v in cls.labeled_outputs.values() for lf in v]
        # all_labels_ = [{k:lf.label} for k, v in cls.labeled_outputs.items() for lf in v]
        all_flows = [lf for v in cls.labeled_outputs.values() for lf in v]
        # Now filter out only those labels that appear more than once
        label_count = Counter(all_labels)
        multi_case_labels = {label for label, count in label_count.items() if count > 1}
        # For each multi_case_label, verify type/content/multipicity compatibility
        # We create a dict keyed by multi_case_label wiht a list of flows per key
        mcl_flows = {mcl: [f.flow for f in all_flows if f.label == mcl] for mcl in multi_case_labels}
        for flows in mcl_flows.values():
            # Create list of flows from flow component of each Labeled_Flow and verify that they are all compatible
            if not Flow.compatible(flows):
                raise ActionException
            pass
        pass
        # Gate Action
        # Create an output Data Flow for each multi case label
        output_flows = {}

        # Populate the Gate Action and get an action id
        gate_aid = Action.populate(tr=tr_Switch, anum=anum, domain=domain, action_type="gate")
        for mcl in multi_case_labels:
            output_flows[mcl] = Flow.populate_switch_output(label=mcl, ref_flow=mcl_flows[mcl][0],
                                                            anum=anum, domain=domain)
            Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Gate Action', tuples=[
                Gate_Action_i(ID=gate_aid, Output_flow=output_flows[mcl].fid, Activity=anum, Domain=domain)
        ])
        for label, flows in mcl_flows.items():
            for f in flows:
                Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Gate Input', tuples=[
                    Gate_Input_i(Input_flow=f.fid, Activity=anum, Domain=domain,
                                 Gate_action=gate_aid)
                ])

        # for sc, lfset in cls.labeled_outputs.items():
        #     for lf in lfset:
        #         if lf.label in multi_case_labels:
        #             # Create the output flow
        #             Relvar.insert(db=mmdb, tr=tr_Switch, relvar='Data_Flow', tuples=[
        #                 Table_i()
        #             ])
        #             pass

        # Create the output dataflow
        # Create a switch for each such multi_case_label
        # Create an input to the switch for each case where the label appears that connects to
        # and connect the corresponding flow

        # x = {i.label for f in cls.labeled_outputs.items() for i in f}
        # TODO: Create data flow switches
        Transaction.execute(db=mmdb, name=tr_Switch)

        # For a switch statement, the switch action is both the initial and output action
        # Initial, because the Switch Action is the one Action in the statement that does not
        # depend on any other data input.
        # Also the final output since regardless of what case
        return Boundary_Actions(ain={action_id}, aout=cls.output_actions)



"""
restrict_action.py – Populate a Restrict Action instance in PyRAL
"""

import logging
from typing import Set
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.aparse_types import Flow_ap, Activity_ap
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.actions.expressions.restriction_condition import RestrictCondition
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.mmclass_nt import (Relational_Action_i, Table_Action_i, Restrict_Action_i,
                                               Table_Restriction_Condition_i)
from pyral.relvar import Relvar
from pyral.relation import Relation  # Here for debugging
from pyral.transaction import Transaction
from scrall.parse.visitor import Criteria_Selection_a, Rank_Selection_a

_logger = logging.getLogger(__name__)

# Transactions
tr_Restrict_Action = "Restrict Action"


class RestrictAction:
    """
    Create all relations for a Restrict Action
    """

    def __init__(self, input_relation_flow: Flow_ap, selection_parse: Criteria_Selection_a,
                 activity_data: Activity_ap):
        """
        Populate the Restrict Action

        :param input_relation_flow: The source table flow into this restriction
        :param selection_parse:  The parsed Scrall select action group
        :param activity_data:
        """
        # Save attribute values that we will need when creating the various select subsystem
        # classes
        domain = activity_data.domain
        anum = activity_data.anum

        # Populate the Action superclass instance and obtain its action_id
        Transaction.open(db=mmdb, name=tr_Restrict_Action)
        self.action_id = Action.populate(tr=tr_Restrict_Action, anum=anum, domain=domain, action_type="restrict")

        # Populate the output Table Flow using same Table as input flow
        self.output_relation_flow = Flow.populate_relation_flow_by_reference(ref_flow=input_relation_flow, anum=anum,
                                                                             domain=domain)

        # Walk through the criteria parse tree storing any attributes or input flows
        rcond = RestrictCondition(tr=tr_Restrict_Action, action_id=self.action_id, input_nsflow=input_relation_flow,
                                  selection_parse=selection_parse, activity_data=activity_data)
        self.sflows = rcond.input_scalar_flows
        # The first two return values are relevant only to instance selection (Select Action)
        # Restrict action does not use the returned cardinality since output is always a Table Flow
        # Nor does it use the comparision critieria to test for identifier selection

        Relvar.insert(db=mmdb, tr=tr_Restrict_Action, relvar='Table_Restriction_Condition', tuples=[
            Table_Restriction_Condition_i(Restrict_action=self.action_id, Activity=anum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Restrict_Action, relvar='Relational_Action', tuples=[
            Relational_Action_i(ID=self.action_id, Activity=anum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Restrict_Action, relvar='Table_Action', tuples=[
            Table_Action_i(ID=self.action_id, Activity=anum, Domain=domain,
                           Input_a_flow=input_relation_flow.fid, Output_flow=self.output_relation_flow.fid)
        ])
        Relvar.insert(db=mmdb, tr=tr_Restrict_Action, relvar='Restrict_Action', tuples=[
            Restrict_Action_i(ID=self.action_id, Activity=anum, Domain=domain)
        ])
        # We now have a transaction with all select-action instances, enter into the metamodel db
        Transaction.execute(db=mmdb, name=tr_Restrict_Action)  # Restrict Action



"""
Blueprint Model Repository Populator

"""
import logging
import logging.config
import sys
import argparse
from pathlib import Path
from xuml_populate.system import System
from xuml_populate import version
import atexit

_logpath = Path("modeldb.log")
_progname = 'Blueprint model repository populator'

def clean_up():
    """Normal and exception exit activities"""
    _logpath.unlink(missing_ok=True)


def get_logger():
    """Initiate the logger"""
    log_conf_path = Path(__file__).parent / 'log.conf'  # Logging configuration is in this file
    logging.config.fileConfig(fname=log_conf_path, disable_existing_loggers=False)
    return logging.getLogger(__name__)  # Create a logger for this module


# Configure the expected parameters and actions for the argparse module
def parse(cl_input):
    parser = argparse.ArgumentParser(description=_progname)
    parser.add_argument('-s', '--system', action='store',
                        help='Name of the system package')
    parser.add_argument('-D', '--debug', action='store_true',
                        help='Debug mode'),
    parser.add_argument('-L', '--log', action='store_true',
                        help='Generate a diagnostic log file')
    parser.add_argument('-A', '--actions', action='store_true',
                        help='Parse actions'),
    parser.add_argument('-V', '--version', action='store_true',
                        help='Print the current version of the repo populator')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Verbose messages')
    return parser.parse_args(cl_input)


def main():
    # Start logging
    logger = get_logger()
    logger.info(f'{_progname} version: {version}')

    # Parse the command line args
    args = parse(sys.argv[1:])

    if args.version:
        # Just print the version and quit
        print(f'{_progname} version: {version}')
        sys.exit(0)

    if not args.log:
        # If no log file is requested, remove the log file before termination
        atexit.register(clean_up)

    # System package specified
    if args.system:
        system_pkg_path = Path(args.system).resolve()
        s = System(name=system_pkg_path.stem, system_path=system_pkg_path, parse_actions=args.actions,
                   verbose=args.verbose)

    logger.info("No problemo")  # We didn't die on an exception, basically
    if args.verbose:
        print("\nNo problemo")


if __name__ == "__main__":
    main()



"""
flow.py – Populate a Flow in PyRAL
"""

import logging
from xuml_populate.config import mmdb
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction
from typing import Optional, Set, List, Dict
from xuml_populate.populate.actions.table import Table
from xuml_populate.exceptions.action_exceptions import FlowException, ControlFlowHasNoTargetActions
from xuml_populate.populate.mmclass_nt import Data_Flow_i, Flow_i, \
    Multiple_Instance_Flow_i, Single_Instance_Flow_i, Instance_Flow_i, \
    Control_Flow_i, Non_Scalar_Flow_i, Scalar_Flow_i, Relation_Flow_i, Labeled_Flow_i, Unlabeled_Flow_i, \
    Tuple_Flow_i, Table_Flow_i, Control_Dependency_i
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content

# TODO: Add Table and Control Flow population

_logger = logging.getLogger(__name__)

# Transaction
tr_Inst_Flow = "Instance Flow"
tr_Rel_Flow = "Relation Flow"
tr_Scalar_Flow = "Scalar Flow"


class Flow:
    """
    Populate relevant Flow relvars
    """
    flow_id_ctr = {}  # Manage flow id numbering per anum:domain

    @classmethod
    def populate_switch_output(cls, label: str, ref_flow: Flow_ap, anum: str, domain: str) -> Flow_ap:
        """
        The output of a Data Flow Switch is a new Data Flow that matches all properties of any Data Flow providing
        input to the Data Flow Switch.  (all inputs have identical properties, so the ref_flow can be any one of the
        input flows)

        :param label: Switch output flows are always labeled
        :param ref_flow: The reference flow is any of the inputs to the switch
        :param anum: The activity number
        :param domain: The domain name
        :return: A summary of the newly populated switch output Data Flow
        """
        match ref_flow.content:
            case Content.INSTANCE:
                return cls.populate_instance_flow(cname=ref_flow.tname, anum=anum, domain=domain, label=label,
                                                  single=(ref_flow.max_mult == MaxMult.ONE))
            case Content.RELATION:
                return cls.populate_relation_flow_by_reference(ref_flow=ref_flow, anum=anum, domain=domain, label=label)
            case Content.SCALAR:
                return cls.populate_scalar_flow(scalar_type=ref_flow.tname, anum=anum, domain=domain, label=label)
            case _:
                raise FlowException

    @classmethod
    def compatible(cls, flows: List[Flow_ap]) -> bool:
        """
        Ensure that all flows in the given list have the same content, type, and multiplicity.
        If so, return true

        :param flows:
        :return: True if all flows are compatible
        """
        assert len(flows) > 1
        # Set reference properties
        c = flows[0].content
        t = flows[0].tname
        m = flows[0].max_mult
        for f in flows[1:]:
            # Compare to reference properties
            if f.content != c or f.tname != t or f.max_mult != m:
                return False
        return True

    @classmethod
    def find_labeled_ns_flow(cls, name: str, anum: str, domain: str) -> Optional[Flow_ap]:
        """
        Given a name, find a Non Scalar Flow labeled with that name and return its summary.

        Nothing is returned if there is no such flow.

        :param name: A name that should match a Labeled Flow
        :param anum: The activity number
        :param domain: The domain name
        :return:
        """
        fid = cls.find_labeled_flow(name=name, anum=anum, domain=domain)
        if not fid:
            return None
        # Is it an Instance Flow?
        R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
        if_result = Relation.restrict(mmdb, relation='Instance_Flow', restriction=R)
        if if_result.body:
            # Okay, it's an instance flow. Now we need the multiplicity on that flow
            ctype = if_result.body[0]['Class']
            many_if_result = Relation.restrict(mmdb, relation='Multiple_Instance_Flow', restriction=R)
            m = MaxMult.MANY if many_if_result.body else MaxMult.ONE
            return Flow_ap(fid=fid, content=Content.INSTANCE, tname=ctype, max_mult=m)

        # Is it a Relation Flow?
        R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
        rf_result = Relation.restrict(mmdb, relation='Relation_Flow', restriction=R)
        if rf_result.body:
            # It's a Relation Flow. Tuple or Table?
            ttype = rf_result.body[0]['Type']
            ntuples = Relation.restrict(mmdb, relation='Table_Flow', restriction=R)
            m = MaxMult.MANY if ntuples.body else MaxMult.ONE
            return Flow_ap(fid=fid, content=Content.RELATION, tname=ttype, max_mult=m)

    @classmethod
    def find_labeled_scalar_flow(cls, name: str, anum: str, domain: str) -> Optional[Flow_ap]:
        """
        Given a name, find a Scalar Flow labeled with that name and return its summary.

        Nothing is returned if there is no such flow.

        :param name: Name that matches a Labled Flow
        :param anum: The activity number
        :param domain: The domain name
        :return: A flow summary or None if no such labeled flow is defined
        """
        fid = cls.find_labeled_flow(name=name, anum=anum, domain=domain)
        if not fid:
            return None
        R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Scalar_Flow', restriction=R)
        if result.body:
            return Flow_ap(fid=fid, content=Content.SCALAR, tname=result.body[0]['Type'], max_mult=None)
        else:
            return None

    @classmethod
    def find_labeled_flow(cls, name: str, anum: str, domain: str) -> Optional[str]:
        """
        Return the fid of a Labeled Flow if it is defined

        :param name: Flow Name
        :param anum: Activity number
        :param domain: Domain name
        :return: flow id (fid) or None if not found
        """
        R = f"Name:<{name}>, Activity:<{anum}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Labeled_Flow', restriction=R)
        if not result.body:
            return None
        else:
            return result.body[0]['ID']

    @classmethod
    def populate_control_flow(cls, tr: str, enabled_actions: Set[str], anum: str, domain: str,
                              label: Optional[str] = None) -> str:
        """
        Populate a new Control Flow.

        Since a Control Flow must feed an Action, it populates inside the provided transaction
        and does not open and close its own like Data Flows.

        :return: The flow id
        """
        flow_id = cls.populate_flow(tr=tr, anum=anum, domain=domain, label=label)
        Relvar.insert(mmdb, tr=tr, relvar='Control_Flow', tuples=[
            Control_Flow_i(ID=flow_id, Activity=anum, Domain=domain)
        ])
        # Populate one or more targets of each Control Flow
        if len(enabled_actions) < 1:
            _logger.error(f"Control flow requires at least one target action")
            raise ControlFlowHasNoTargetActions
        for a in enabled_actions:
            Relvar.insert(mmdb, tr=tr, relvar='Control_Dependency', tuples=[
                Control_Dependency_i(Control_flow=flow_id, Action=a, Activity=anum, Domain=domain)
            ])
        return flow_id

    @classmethod
    def populate_table_flow_from_class(cls, cname: str, anum: str, domain: str) -> Flow_ap:
        """

        :return:
        """
        pass

    @classmethod
    def lookup_data(cls, fid: str, anum: str, domain: str) -> Flow_ap:
        """
        Given a Data Flow identifier (fid, anum, domain), determine the flow content, tname, and maxmult

        Raise exception if no such Data Flow

        :param fid: Flow ID of a Data Flow
        :param anum: The activity number
        :param domain: The domain name
        :return: A flow summary for the supplied ID
        """
        # First verify that the fid corresponds to some Data Flow instance
        R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Data_Flow', restriction=R)
        if not result.body:
            # Either fid not defined or it is a Control Flow
            raise FlowException

        # Is Non Scalar or Scalar Flow?
        R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Non_Scalar_Flow', restriction=R)
        if result.body:
            # It is a Non Scalar Flow
            R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
            result = Relation.restrict(mmdb, relation='Instance_Flow', restriction=R)
            if result.body:
                # It's an Instance Flow
                tname = result.body[0]['Type']
                R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
                result = Relation.restrict(mmdb, relation='Multiple_Instance_Flow', restriction=R)
                max_mult = MaxMult.MANY if result.body else MaxMult.ONE
                return Flow_ap(fid=fid, content=Content.INSTANCE, tname=tname, max_mult=max_mult)
            else:
                # Must be a Table Flow
                R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
                result = Relation.restrict(mmdb, relation='Relation_Flow', restriction=R)
                tname = result.body[0]['Type']
                R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
                result = Relation.restrict(mmdb, relation='Table_Flow', restriction=R)
                max_mult = MaxMult.MANY if result.body else MaxMult.ONE
                return Flow_ap(fid=fid, content=Content.RELATION, tname=tname, max_mult=max_mult)
        else:
            # It's a Scalar Flow
            R = f"ID:<{fid}>, Activity:<{anum}>, Domain:<{domain}>"
            result = Relation.restrict(mmdb, relation='Scalar_Flow', restriction=R)
            tname = result.body[0]['Type']
            return Flow_ap(fid=fid, content=Content.SCALAR, tname=tname, max_mult=None)

    @classmethod
    def populate_data_flow_by_type(cls, label: Optional[str], mm_type: str, anum: str,
                                   domain: str, activity_tr: str = None) -> Flow_ap:
        """
        Populate an instance of Data Flow and determine its subclasses based on the supplied
        Class, Scalar, or Table Type.

        :param label:
        :param mm_type: A Class, Scalar, or Table Type
        :param anum:
        :param domain:
        :param activity_tr: Incorporate this in the activity population transaction if provided
        :return: The generated flow id
        """
        tr = tr_Inst_Flow if not activity_tr else activity_tr
        # For now we distinguish only between class and scalar types
        # Is the type a Class Type?
        R = f"Name:<{mm_type}>, Domain:<{domain}>"
        r_result = Relation.restrict(db=mmdb, relation='Class', restriction=R)
        if r_result.body:
            # It's a class type, create a multiple instance flow
            flow = cls.populate_instance_flow(cname=mm_type, anum=anum, domain=domain, label=label,
                                              activity_tr=tr)
        else:
            # It's a scalar type
            flow = cls.populate_scalar_flow(scalar_type=mm_type, anum=anum, domain=domain,
                                            label=label, activity_tr=tr)
        return flow

    @classmethod
    def populate_scalar_flow(cls, scalar_type: str, anum: str, domain: str, label: Optional[str] = None,
                             activity_tr: str = None) -> Flow_ap:
        """
        Populate an instance of Scalar flow

        :param scalar_type: The name of the Scalar (Type)
        :param anum: The anum of the enclosing Activity
        :param domain: The name of the domain
        :param label: If provided, a labeled flow is populated
        :param activity_tr: Incorporate this in the activity population transaction if provided
        :return: A Flow_ap summary of the key flow characteristics
        """
        tr = tr_Inst_Flow if not activity_tr else activity_tr
        # See comment in populate_instance_flow
        if not activity_tr:
            Transaction.open(db=mmdb, name=tr)

        flow_id = cls.populate_data_flow(tr=tr, anum=anum, domain=domain, label=label)
        Relvar.insert(db=mmdb, tr=tr, relvar='Scalar_Flow', tuples=[
            Scalar_Flow_i(ID=flow_id, Activity=anum, Domain=domain, Type=scalar_type)
        ])
        if not activity_tr:
            Transaction.execute(db=mmdb, name=tr)

        return Flow_ap(fid=flow_id, content=Content.SCALAR, tname=scalar_type, max_mult=None)

    @classmethod
    def populate_instance_flow(cls, cname: str, anum: str, domain: str, label: Optional[str] = None,
                               single: bool = False, activity_tr: str = None) -> Flow_ap:
        """
        Populate an instance of Scalar flow

        :param cname: The class name which establishes the Type
        :param label: If provided, a labeled flow is populated
        :param anum: The anum of the enclosing Activity
        :param domain: The name of the domain
        :param single: If true, Single vs Multiple Instance Flow which is assumed by default
        :param activity_tr: Incorporate this in the activity population transaction if provided
        :return: A Flow_ap summary of the key flow characteristics
        """
        # When an Activity is being populated, it may require certain Flows, the 'me' initial
        # executing instance Flow, for example. In this case, the Activity has a transaction open
        # and, until this outer transaction closes, the Activity does not yet exist. In this case,
        # we use the activity transaction, otherwise we create a new one locally
        tr = tr_Inst_Flow if not activity_tr else activity_tr
        if not activity_tr:
            Transaction.open(db=mmdb, name=tr)

        flow_id = cls.populate_non_scalar_flow(tr=tr, anum=anum, domain=domain, label=label)
        Relvar.insert(db=mmdb, tr=tr, relvar='Instance_Flow', tuples=[
            Instance_Flow_i(ID=flow_id, Activity=anum, Domain=domain, Class=cname)
        ])
        if single:
            max_mult = MaxMult.ONE
            Relvar.insert(db=mmdb, tr=tr, relvar='Single_Instance_Flow', tuples=[
                Single_Instance_Flow_i(ID=flow_id, Activity=anum, Domain=domain)
            ])
        else:
            max_mult = MaxMult.MANY
            Relvar.insert(db=mmdb, tr=tr, relvar='Multiple_Instance_Flow', tuples=[
                Multiple_Instance_Flow_i(ID=flow_id, Activity=anum, Domain=domain)
            ])

        if not activity_tr:
            Transaction.execute(db=mmdb, name=tr)

        return Flow_ap(fid=flow_id, content=Content.INSTANCE, tname=cname, max_mult=max_mult)

    @classmethod
    def populate_non_scalar_flow(cls, tr: str, anum: str, domain: str, label: Optional[str] = None) -> str:
        """
        Populate an instance of Non Scalar flow (called by subclass only)

        :param tr:  The transaction specific to Flow subclass
        :param anum: The activity number
        :param domain: The domain name
        :param label: An optional label
        :return: The Flow ID
        """
        fid = cls.populate_data_flow(tr=tr, anum=anum, domain=domain, label=label)
        Relvar.insert(db=mmdb, tr=tr, relvar='Non_Scalar_Flow', tuples=[
            Non_Scalar_Flow_i(ID=fid, Activity=anum, Domain=domain)
        ])
        return fid

    @classmethod
    def populate_data_flow(cls, tr: str, anum: str, domain: str, label: Optional[str] = None) -> str:
        """
        """
        fid = cls.populate_flow(tr=tr, anum=anum, domain=domain, label=label)
        Relvar.insert(db=mmdb, tr=tr, relvar='Data_Flow', tuples=[
            Data_Flow_i(ID=fid, Activity=anum, Domain=domain)
        ])
        return fid

    @classmethod
    def populate_relation_flow(cls, table_name: str, anum: str, domain: str, is_tuple: bool,
                               label: Optional[str] = None) -> Flow_ap:
        """
        Given an existing Table, populate a Relation Flow

        :param table_name: Name of an existing Table
        :param anum:
        :param domain:
        :param is_tuple:
        :param label:
        :return: Summary of the populated flow
        """
        flow_id = cls.populate_non_scalar_flow(tr=tr_Rel_Flow, anum=anum, domain=domain, label=label)
        Relvar.insert(db=mmdb, tr=tr_Rel_Flow, relvar='Relation_Flow', tuples=[
            Relation_Flow_i(ID=flow_id, Activity=anum, Domain=domain, Type=table_name)
        ])
        if is_tuple:
            Relvar.insert(db=mmdb, tr=tr_Rel_Flow, relvar='Tuple_Flow', tuples=[
                Tuple_Flow_i(ID=flow_id, Activity=anum, Domain=domain)
            ])
        else:
            Relvar.insert(db=mmdb, tr=tr_Rel_Flow, relvar='Table_Flow', tuples=[
                Table_Flow_i(ID=flow_id, Activity=anum, Domain=domain)
            ])
        return Flow_ap(fid=flow_id, content=Content.RELATION, tname=table_name,
                       max_mult=MaxMult.ONE if is_tuple else MaxMult.MANY)

    @classmethod
    def populate_relation_flow_by_reference(cls, ref_flow: Flow_ap, anum: str, domain: str, tuple_flow: bool = False,
                                            label: Optional[str] = None) -> Flow_ap:
        """
        Given an existing Relation Flow, create another with the same properties.

        This is useful in the Switch Action for creating the output of a Data Flow Switch which must match
        the input.

        :param ref_flow: A Relation Flow summary
        :param label: optional label
        :param anum: The anum number
        :param domain: The domain name
        :param tuple_flow:  If True, disregard the reference cardinality and force a tuple result
        :return: Summary of duplicated flow
        """
        # Since we are making a copy, we know that the Table (Type) must already exist, so we skip the Table creation
        # / verification step and just populate the Relation Flow.
        # TODO: Log this
        Transaction.open(db=mmdb, name=tr_Rel_Flow)
        set_tuple = True if tuple_flow else ref_flow.max_mult == MaxMult.ONE
        rflow = cls.populate_relation_flow(table_name=ref_flow.tname, anum=anum, domain=domain,
                                           is_tuple=set_tuple, label=label)
        Transaction.execute(db=mmdb, name=tr_Rel_Flow)
        return rflow

    @classmethod
    def populate_relation_flow_by_header(cls, table_header: Dict[str, str], anum: str, domain: str, max_mult: MaxMult,
                                         label: Optional[str] = None) -> Flow_ap:
        """
        Populate a Relation Flow as either a Table or Tuple Flow typed with either an existing Table,
        if there is a header match, or a newly created Table.

        :param table_header: The table header as a dict of attribute name;type pairs
        :param anum: Activity number (anum)
        :param domain: The domain name
        :param max_mult: Max number of tuples
        :param label: Optional name of flow
        :return: The newly populated Relation Flow
        """
        Transaction.open(mmdb, tr_Rel_Flow)
        # Populate the table if it does not aleady exist
        table_name = Table.populate(tr=tr_Rel_Flow, table_header=table_header, domain=domain)
        rflow = cls.populate_relation_flow(table_name=table_name, anum=anum, domain=domain,
                                           is_tuple=(max_mult == MaxMult.ONE), label=label)
        Transaction.execute(mmdb, tr_Rel_Flow)
        return rflow

    @classmethod
    def populate_flow(cls, tr: str, anum: str, domain: str, label: Optional[str] = None) -> str:
        """
        Called only by a subclass Flow populator, populate a Flow of the subclass with an optional label.

        :param tr: The open transaction specific to the Flow subclass
        :param anum: The anum
        :param domain: The domain name
        :param label: An optional label
        :return: The Flow ID (fid)
        """
        """
        Populate Flow instance and optional Label
        """
        # Each anum requires a new flow id counter
        activity_id = f'{domain}:{anum}'  # combine attributes to get id
        if activity_id not in cls.flow_id_ctr.keys():
            cls.flow_id_ctr[activity_id] = 0

        # Populate Flow instance
        cls.flow_id_ctr[activity_id] += 1  # Increment the flow id counter for this anum
        fid = f"F{cls.flow_id_ctr[activity_id]}"
        Relvar.insert(db=mmdb, tr=tr, relvar='Flow', tuples=[
            Flow_i(ID=fid, Activity=anum, Domain=domain)
        ])

        # If a label has been defined, populate it
        if label:
            Relvar.insert(db=mmdb, tr=tr, relvar='Labeled_Flow', tuples=[
                Labeled_Flow_i(Name=label, ID=fid, Activity=anum, Domain=domain)
            ])
        else:
            Relvar.insert(db=mmdb, tr=tr, relvar='Unlabeled_Flow', tuples=[
                Unlabeled_Flow_i(ID=fid, Activity=anum, Domain=domain)
            ])

        return fid
